#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# soda.pyx
#
# Cython file defining the thick implementation SODA classes (embedded in
# thick_impl.pyx).
#------------------------------------------------------------------------------

cdef class ThickSodaDbImpl(BaseSodaDbImpl):
    cdef:
        dpiSodaDb* _handle

    def __dealloc__(self):
        if self._handle != NULL:
            dpiSodaDb_release(self._handle)

    cdef int _get_flags(self, uint32_t* flags) except -1:
        self._conn._verify_connected()
        if self._conn.autocommit:
            flags[0] = DPI_SODA_FLAGS_ATOMIC_COMMIT
        else:
            flags[0] = DPI_SODA_FLAGS_DEFAULT

    def create_collection(self, str name, str metadata, bint map_mode):
        """
        Internal method for creating a collection.
        """
        cdef:
            StringBuffer metadata_buf = StringBuffer()
            StringBuffer name_buf = StringBuffer()
            ThickSodaCollImpl coll_impl
            uint32_t flags
            int status
        name_buf.set_value(name)
        metadata_buf.set_value(metadata)
        self._get_flags(&flags)
        if map_mode:
            flags |= DPI_SODA_FLAGS_CREATE_COLL_MAP
        coll_impl = ThickSodaCollImpl.__new__(ThickSodaCollImpl)
        coll_impl._db_impl = self
        with nogil:
            status = dpiSodaDb_createCollection(self._handle, name_buf.ptr,
                                                name_buf.length,
                                                metadata_buf.ptr,
                                                metadata_buf.length, flags,
                                                &coll_impl._handle)
        if status < 0:
            _raise_from_odpi()
        coll_impl._get_name()
        return coll_impl

    def create_document(self, bytes content, str key, str media_type):
        """
        Internal method for creating a document.
        """
        cdef:
            StringBuffer media_type_buf = StringBuffer()
            StringBuffer content_buf = StringBuffer()
            StringBuffer key_buf = StringBuffer()
            ThickSodaDocImpl doc_impl
        content_buf.set_value(content)
        key_buf.set_value(key)
        media_type_buf.set_value(media_type)
        doc_impl = ThickSodaDocImpl.__new__(ThickSodaDocImpl)
        if dpiSodaDb_createDocument(self._handle, key_buf.ptr, key_buf.length,
                                    content_buf.ptr, content_buf.length,
                                    media_type_buf.ptr, media_type_buf.length,
                                    DPI_SODA_FLAGS_DEFAULT,
                                    &doc_impl._handle) < 0:
            _raise_from_odpi()
        return doc_impl

    def get_collection_names(self, str start_name, uint32_t limit):
        """
        Internal method for getting the list of collection names.
        """
        cdef:
            StringBuffer start_name_buf = StringBuffer()
            dpiSodaCollNames coll_names
            uint32_t flags, i
            list result
            int status
            str temp
        start_name_buf.set_value(start_name)
        self._get_flags(&flags)
        with nogil:
            status = dpiSodaDb_getCollectionNames(self._handle,
                                                  start_name_buf.ptr,
                                                  start_name_buf.length,
                                                  limit, flags, &coll_names)
        if status < 0:
            _raise_from_odpi()
        try:
            result = cpython.PyList_New(coll_names.numNames)
            for i in range(coll_names.numNames):
                temp = coll_names.names[i][:coll_names.nameLengths[i]].decode()
                cpython.Py_INCREF(temp)
                cpython.PyList_SET_ITEM(result, i, temp)
            return result
        finally:
            if dpiSodaDb_freeCollectionNames(self._handle, &coll_names) < 0:
                _raise_from_odpi()

    def open_collection(self, str name):
        """
        Internal method for opening a collection.
        """
        cdef:
            StringBuffer name_buf = StringBuffer()
            ThickSodaCollImpl coll_impl
            uint32_t flags
            int status
        name_buf.set_value(name)
        self._get_flags(&flags)
        coll_impl = ThickSodaCollImpl.__new__(ThickSodaCollImpl)
        coll_impl._db_impl = self
        with nogil:
            status = dpiSodaDb_openCollection(self._handle, name_buf.ptr,
                                              name_buf.length, flags,
                                              &coll_impl._handle)
        if status < 0:
            _raise_from_odpi()
        if coll_impl._handle != NULL:
            coll_impl._get_name()
            return coll_impl


cdef class ThickSodaCollImpl(BaseSodaCollImpl):
    cdef:
        ThickSodaDbImpl _db_impl
        dpiSodaColl* _handle

    def __dealloc__(self):
        if self._handle != NULL:
            dpiSodaColl_release(self._handle)

    cdef int _get_name(self) except -1:
        """
        Internal method for getting the name of the collection.
        """
        cdef:
            uint32_t name_len
            const char *name
        if dpiSodaColl_getName(self._handle, &name, &name_len) < 0:
            _raise_from_odpi()
        self.name = name[:name_len].decode()

    cdef int _process_options(self, dpiSodaOperOptions *options,
                              const char *ptr, uint32_t length) except -1:
        """
        Internal method for populating the SODA operations structure with the
        information provided by the user.
        """
        if dpiContext_initSodaOperOptions(driver_context, options) < 0:
            _raise_from_odpi()
        options.hint = ptr
        options.hintLength = length

    def create_index(self, str spec):
        """
        Internal method for creating an index on a collection.
        """
        cdef:
            StringBuffer buf = StringBuffer()
            uint32_t flags
            int status
        buf.set_value(spec)
        self._db_impl._get_flags(&flags)
        with nogil:
            status = dpiSodaColl_createIndex(self._handle, buf.ptr, buf.length,
                                             flags)
        if status < 0:
            _raise_from_odpi()

    def drop(self):
        """
        Internal method for dropping a collection.
        """
        cdef:
            bint is_dropped
            uint32_t flags
        self._db_impl._get_flags(&flags)
        if dpiSodaColl_drop(self._handle, flags, &is_dropped) < 0:
            _raise_from_odpi()
        return is_dropped

    def drop_index(self, str name, bint force):
        """
        Internal method for dropping an index on a collection.
        """
        cdef:
            StringBuffer buf = StringBuffer()
            bint is_dropped
            uint32_t flags
            int status
        buf.set_value(name)
        self._db_impl._get_flags(&flags)
        if force:
            flags |= DPI_SODA_FLAGS_INDEX_DROP_FORCE
        with nogil:
            status = dpiSodaColl_dropIndex(self._handle, buf.ptr, buf.length,
                                           flags, &is_dropped)
        if status < 0:
            _raise_from_odpi()
        return is_dropped

    def get_count(self, object op):
        """
        Internal method for getting the count of documents matching the
        criteria.
        """
        cdef:
            ThickSodaOpImpl options = ThickSodaOpImpl._from_op(op)
            uint64_t count
            uint32_t flags
            int status
        self._db_impl._get_flags(&flags)
        with nogil:
            status = dpiSodaColl_getDocCount(self._handle, &options._options,
                                             flags, &count)
        if status < 0:
            _raise_from_odpi()
        return count

    def get_cursor(self, object op):
        """
        Internal method for getting a cursor which will return the documents
        matching the criteria.
        """
        cdef:
            ThickSodaOpImpl options = ThickSodaOpImpl._from_op(op)
            ThickSodaDocCursorImpl cursor_impl
            uint32_t flags
            int status
        self._db_impl._get_flags(&flags)
        cursor_impl = ThickSodaDocCursorImpl.__new__(ThickSodaDocCursorImpl)
        cursor_impl._db_impl = self._db_impl
        with nogil:
            status = dpiSodaColl_find(self._handle, &options._options, flags,
                                      &cursor_impl._handle)
        if status < 0:
            _raise_from_odpi()
        return cursor_impl

    def get_data_guide(self):
        """
        Internal method for getting the data guide for a collection.
        """
        cdef:
            ThickSodaDocImpl doc_impl
            uint32_t flags
            int status
        self._db_impl._get_flags(&flags)
        doc_impl = ThickSodaDocImpl.__new__(ThickSodaDocImpl)
        with nogil:
            status = dpiSodaColl_getDataGuide(self._handle, flags,
                                              &doc_impl._handle)
        if status < 0:
            _raise_from_odpi()
        if doc_impl._handle != NULL:
            return doc_impl

    def get_metadata(self):
        """
        Internal method for getting the metadata for a collection.
        """
        cdef:
            uint32_t value_len
            const char* value
        if dpiSodaColl_getMetadata(self._handle, &value, &value_len) < 0:
            _raise_from_odpi()
        return value[:value_len].decode()

    def get_one(self, object op):
        """
        Internal method for getting a document matching the criteria.
        """
        cdef:
            ThickSodaOpImpl options = ThickSodaOpImpl._from_op(op)
            ThickSodaDocImpl doc_impl
            uint32_t flags
            int status
        self._db_impl._get_flags(&flags)
        doc_impl = ThickSodaDocImpl.__new__(ThickSodaDocImpl)
        with nogil:
            status = dpiSodaColl_findOne(self._handle, &options._options,
                                         flags, &doc_impl._handle)
        if status < 0:
            _raise_from_odpi()
        if doc_impl._handle != NULL:
            return doc_impl

    def insert_many(self, list doc_impls, str hint, bint return_docs):
        """
        Internal method for inserting many documents into a collection at once.
        """
        cdef:
            dpiSodaDoc **output_handles = NULL
            uint32_t i, num_docs, flags
            ThickSodaDocImpl doc_impl
            list output_doc_impls
            dpiSodaDoc **handles
            ssize_t num_bytes
            dpiSodaOperOptions options
            dpiSodaOperOptions *options_ptr = NULL
            StringBuffer hint_buf = StringBuffer()
            int status
        num_docs = <uint32_t> len(doc_impls)
        num_bytes = num_docs * sizeof(dpiSodaDoc *)
        handles = <dpiSodaDoc**> cpython.PyMem_Malloc(num_bytes)
        if return_docs:
            output_handles = <dpiSodaDoc**> _calloc(num_docs,
                                                    sizeof(dpiSodaDoc*))
            if hint is not None:
                hint_buf.set_value(hint)
                options_ptr = &options
                self._process_options(&options, hint_buf.ptr, hint_buf.length)
        for i, doc_impl in enumerate(doc_impls):
            handles[i] = doc_impl._handle
        self._db_impl._get_flags(&flags)
        with nogil:
            status = dpiSodaColl_insertManyWithOptions(self._handle, num_docs,
                                                       handles, options_ptr,
                                                       flags, output_handles)
        if status < 0:
            _raise_from_odpi()
        if return_docs:
            output_doc_impls = []
            for i in range(num_docs):
                doc_impl = ThickSodaDocImpl.__new__(ThickSodaDocImpl)
                doc_impl._handle = output_handles[i]
                output_doc_impls.append(doc_impl)
            return output_doc_impls

    def insert_one(self, ThickSodaDocImpl doc_impl, str hint, bint return_doc):
        """
        Internal method for inserting a single document into a collection.
        """
        cdef:
            dpiSodaDoc **output_handle = NULL
            ThickSodaDocImpl output_doc_impl
            uint32_t flags
            dpiSodaOperOptions options
            dpiSodaOperOptions *options_ptr = NULL
            StringBuffer hint_buf = StringBuffer()
            int status
        if return_doc:
            output_doc_impl = ThickSodaDocImpl.__new__(ThickSodaDocImpl)
            output_handle = &output_doc_impl._handle
            if hint is not None:
               hint_buf.set_value(hint)
               options_ptr = &options
               self._process_options(&options, hint_buf.ptr, hint_buf.length)
        self._db_impl._get_flags(&flags)
        with nogil:
            status = dpiSodaColl_insertOneWithOptions(self._handle,
                                                      doc_impl._handle,
                                                      options_ptr, flags,
                                                      output_handle)
        if status < 0:
            _raise_from_odpi()
        if return_doc:
            return output_doc_impl

    def remove(self, object op):
        """
        Internal method for removing all of the documents matching the
        criteria.
        """
        cdef:
            ThickSodaOpImpl options = ThickSodaOpImpl._from_op(op)
            uint64_t count
            uint32_t flags
            int status
        self._db_impl._get_flags(&flags)
        with nogil:
            status = dpiSodaColl_remove(self._handle, &options._options, flags,
                                        &count)
        if status < 0:
            _raise_from_odpi()
        return count

    def replace_one(self, object op, ThickSodaDocImpl doc_impl,
                    bint return_doc):
        """
        Internal method for replacing the document matching the criteria with
        the supplied coument.
        """
        cdef:
            ThickSodaOpImpl options = ThickSodaOpImpl._from_op(op)
            dpiSodaDoc **output_handle = NULL
            ThickSodaDocImpl output_doc_impl
            uint32_t flags
            bint replaced
            int status
        if return_doc:
            output_doc_impl = ThickSodaDocImpl.__new__(ThickSodaDocImpl)
            output_handle = &output_doc_impl._handle
        self._db_impl._get_flags(&flags)
        with nogil:
            status = dpiSodaColl_replaceOne(self._handle, &options._options,
                                            doc_impl._handle, flags, &replaced,
                                            output_handle)
        if status < 0:
            _raise_from_odpi()
        if return_doc:
            return output_doc_impl

    def save(self, ThickSodaDocImpl doc_impl, str hint, bint return_doc):
        """
        Internal method for saving a document into the collection.
        """
        cdef:
            dpiSodaDoc **output_handle = NULL
            ThickSodaDocImpl output_doc_impl
            uint32_t flags
            dpiSodaOperOptions options
            dpiSodaOperOptions *options_ptr = NULL
            StringBuffer hint_buf = StringBuffer()
            int status
        if return_doc:
            output_doc_impl = ThickSodaDocImpl.__new__(ThickSodaDocImpl)
            output_handle = &output_doc_impl._handle
            if hint is not None:
               hint_buf.set_value(hint)
               options_ptr = &options
               self._process_options(&options, hint_buf.ptr, hint_buf.length)
        self._db_impl._get_flags(&flags)
        with nogil:
            status = dpiSodaColl_saveWithOptions(self._handle,
                                                 doc_impl._handle,
                                                 options_ptr, flags,
                                                 output_handle)
        if status < 0:
            _raise_from_odpi()
        if return_doc:
            return output_doc_impl

    def truncate(self):
        """
        Internal method for truncating the collection (removing all documents
        from it).
        """
        cdef int status
        with nogil:
            status = dpiSodaColl_truncate(self._handle)
        if status < 0:
            _raise_from_odpi()


cdef class ThickSodaDocImpl(BaseSodaDocImpl):
    cdef:
        dpiSodaDoc* _handle

    def __dealloc__(self):
        if self._handle != NULL:
            dpiSodaDoc_release(self._handle)

    def get_content(self):
        """
        Internal method for returning the content of the document.
        """
        cdef:
            bytes out_content = None
            str out_encoding = None
            const char *encoding
            uint32_t content_len
            const char *content
        if dpiSodaDoc_getContent(self._handle, &content, &content_len,
                                 &encoding) < 0:
            _raise_from_odpi()
        if content != NULL:
            out_content = content[:content_len]
        if encoding != NULL:
            out_encoding = encoding.decode()
        else:
            out_encoding = "UTF-8"
        return (out_content, out_encoding)

    def get_created_on(self):
        """
        Internal method for getting the date the document was created.
        """
        cdef:
            uint32_t value_len
            const char *value
        if dpiSodaDoc_getCreatedOn(self._handle, &value, &value_len) < 0:
            _raise_from_odpi()
        if value_len > 0:
            return value[:value_len].decode()

    def get_key(self):
        """
        Internal method for getting the key of the document.
        """
        cdef:
            uint32_t value_len
            const char *value
        if dpiSodaDoc_getKey(self._handle, &value, &value_len) < 0:
            _raise_from_odpi()
        if value_len > 0:
            return value[:value_len].decode()

    def get_last_modified(self):
        """
        Internal method for getting the date the document was last modified.
        """
        cdef:
            uint32_t value_len
            const char *value
        if dpiSodaDoc_getLastModified(self._handle, &value, &value_len) < 0:
            _raise_from_odpi()
        if value_len > 0:
            return value[:value_len].decode()

    def get_media_type(self):
        """
        Internal method for getting the media type of the document.
        """
        cdef:
            uint32_t value_len
            const char *value
        if dpiSodaDoc_getMediaType(self._handle, &value, &value_len) < 0:
            _raise_from_odpi()
        if value_len > 0:
            return value[:value_len].decode()

    def get_version(self):
        """
        Internal method for getting the version of the document.
        """
        cdef:
            uint32_t value_len
            const char *value
        if dpiSodaDoc_getVersion(self._handle, &value, &value_len) < 0:
            _raise_from_odpi()
        if value_len > 0:
            return value[:value_len].decode()


cdef class ThickSodaDocCursorImpl(BaseSodaDocCursorImpl):
    cdef:
        dpiSodaDocCursor* _handle
        ThickSodaDbImpl _db_impl

    def __dealloc__(self):
        if self._handle != NULL:
            dpiSodaDocCursor_release(self._handle)

    def close(self):
        """
        Internal method for closing the cursor.
        """
        cdef int status
        with nogil:
            status = dpiSodaDocCursor_close(self._handle)
        if status < 0:
            _raise_from_odpi()

    def get_next_doc(self):
        """
        Internal method for getting the next document from the cursor.
        """
        cdef:
            ThickSodaDocImpl doc_impl
            int status
        doc_impl = ThickSodaDocImpl.__new__(ThickSodaDocImpl)
        with nogil:
            status = dpiSodaDocCursor_getNext(self._handle,
                                              DPI_SODA_FLAGS_DEFAULT,
                                              &doc_impl._handle)
        if status < 0:
            _raise_from_odpi()
        if doc_impl._handle != NULL:
            return doc_impl


cdef class ThickSodaOpImpl:
    cdef:
        dpiSodaOperOptions _options
        const char** _key_values
        uint32_t* _key_lengths
        list _buffers

    def __dealloc__(self):
        if self._key_values != NULL:
            cpython.PyMem_Free(self._key_values)
        if self._key_lengths != NULL:
            cpython.PyMem_Free(self._key_lengths)

    cdef int _add_buf(self, object value, const char **ptr,
                      uint32_t *length) except -1:
        cdef StringBuffer buf = StringBuffer()
        buf.set_value(value)
        self._buffers.append(buf)
        ptr[0] = buf.ptr
        length[0] = buf.length

    @staticmethod
    cdef ThickSodaOpImpl _from_op(object op):
        """
        Internal method for creating a SODA operations implementation object
        given the object supplied by the user.
        """
        cdef:
            ThickSodaOpImpl impl = ThickSodaOpImpl.__new__(ThickSodaOpImpl)
            dpiSodaOperOptions *options
            ssize_t num_bytes
            uint32_t i
        impl._buffers = []
        options = &impl._options
        if dpiContext_initSodaOperOptions(driver_context, options) < 0:
            _raise_from_odpi()
        if op._keys:
            options.numKeys = <uint32_t> len(op._keys)
            num_bytes = options.numKeys * sizeof(char*)
            impl._key_values = <const char**> cpython.PyMem_Malloc(num_bytes)
            num_bytes = options.numKeys * sizeof(uint32_t)
            impl._key_lengths = <uint32_t*> cpython.PyMem_Malloc(num_bytes)
            options.keys = impl._key_values
            options.keyLengths = impl._key_lengths
            for i in range(options.numKeys):
                impl._add_buf(op._keys[i], &impl._key_values[i],
                              &impl._key_lengths[i])
        if op._key is not None:
            impl._add_buf(op._key, &options.key, &options.keyLength)
        if op._version is not None:
            impl._add_buf(op._version, &options.version,
                          &options.versionLength)
        if op._filter is not None:
            impl._add_buf(op._filter, &options.filter, &options.filterLength)
        if op._hint is not None:
            impl._add_buf(op._hint, &options.hint, &options.hintLength)
        if op._skip is not None:
            options.skip = op._skip
        if op._limit is not None:
            options.limit = op._limit
        if op._fetch_array_size is not None:
            options.fetchArraySize = op._fetch_array_size
        return impl



#------------------------------------------------------------------------------
# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# driver_mode.py
#
# Contains a simple method for checking and returning which mode the driver is
# currently using. The driver only supports creating connections and pools with
# either the thin implementation or the thick implementation, not both
# simultaneously.
#------------------------------------------------------------------------------

import threading

from . import errors

# The DriverModeHandler class is used to manage which mode the driver is using.
#
# The "thin_mode" flag contains the current state:
#     None: neither thick nor thin implementation has been used yet
#     False: thick implementation is being used
#     True: thin implementation is being used
#
# The "requested_thin_mode" flag is set to the mode that is being requested:
#     False: thick implementation is being initialized
#     True: thin implementation is being initialized
class DriverModeManager:
    """
    Manages the mode the driver is using. The "thin_mode" flag contains the
    current state:
        None: neither thick nor thin implementation has been used yet
        False: thick implementation is being used
        True: thin implementation is being used
    The "requested_thin_mode" is set to the mode that is being requested, but
    only while initialization is taking place (otherwise, it contains the value
    None):
        False: thick implementation is being initialized
        True: thin implementation is being initialized
    The condition is used to ensure that only one thread is performing
    initialization.
    """
    def __init__(self):
        self.thin_mode = None
        self.requested_thin_mode = None
        self.condition = threading.Condition()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        with self.condition:
            if exc_type is None and exc_value is None and exc_tb is None \
                    and self.requested_thin_mode is not None:
                self.thin_mode = self.requested_thin_mode
            self.requested_thin_mode = None
            self.condition.notify()

    @property
    def thin(self):
        if self.requested_thin_mode is not None:
            return self.requested_thin_mode
        return self.thin_mode

manager = DriverModeManager()

def get_manager(requested_thin_mode=None):
    """
    Returns the manager, but only after ensuring that no other threads are
    attempting to initialize the mode.

    NOTE: the current implementation of the driver only requires
    requested_thin_mode to be set when initializing the thick mode; for this
    reason the error raised is specified about a thin mode connection already
    being created. If this assumption changes, a new error message will be
    required.
    """
    with manager.condition:
        if manager.thin_mode is None:
            if manager.requested_thin_mode is not None:
                manager.condition.wait()
            if manager.thin_mode is None:
                if requested_thin_mode is None:
                    manager.requested_thin_mode = True
                else:
                    manager.requested_thin_mode = requested_thin_mode
        elif requested_thin_mode is not None \
                and requested_thin_mode != manager.thin_mode:
            errors._raise_err(errors.ERR_THIN_CONNECTION_ALREADY_CREATED)
    return manager


def is_thin_mode() -> bool:
    """
    Return a boolean specifying whether the driver is using thin mode (True) or
    thick mode (False).

    Immediately after python-oracledb is imported, this function will return
    True indicating that python-oracledb defaults to Thin mode. If
    oracledb.init_oracle_client() is called successfully, then a subsequent
    call to is_thin_mode() will return False indicating that Thick mode is
    enabled.  Once the first standalone connection or connection pool is
    created succesfully, or a call to oracledb.init_oracle_client() is made
    successfully, then python-oracledb's mode is fixed and the value returned
    by is_thin_mode() will never change for the lifetime of the process.

    """
    if manager.thin_mode is not None:
        return manager.thin_mode
    return True



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
2600 - Module for testing timestamp variables
"""

import time

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.raw_data = []
        self.data_by_key = {}
        for i in range(1, 11):
            time_tuple = (2002, 12, 9, 0, 0, 0, 0, 0, -1)
            time_in_ticks = time.mktime(time_tuple) + i * 86400
            date_value = oracledb.TimestampFromTicks(int(time_in_ticks))
            str_value = str(i * 50)
            fsecond = int(str_value + "0" * (6 - len(str_value)))
            date_col = oracledb.Timestamp(date_value.year, date_value.month,
                                          date_value.day, date_value.hour,
                                          date_value.minute, i * 2, fsecond)
            if i % 2:
                time_in_ticks = time.mktime(time_tuple) + i * 86400 + 86400
                date_value = oracledb.TimestampFromTicks(int(time_in_ticks))
                str_value = str(i * 125)
                fsecond = int(str_value + "0" * (6 - len(str_value)))
                nullable_col = oracledb.Timestamp(date_value.year,
                                                  date_value.month,
                                                  date_value.day,
                                                  date_value.hour,
                                                  date_value.minute, i * 3,
                                                  fsecond)
            else:
                nullable_col = None
            data_tuple = (i, date_col, nullable_col)
            self.raw_data.append(data_tuple)
            self.data_by_key[i] = data_tuple

    def test_2600_bind_timestamp(self):
        "2600 - test binding in a timestamp"
        self.cursor.setinputsizes(value=oracledb.DB_TYPE_TIMESTAMP)
        self.cursor.execute("""
                select * from TestTimestamps
                where TimestampCol = :value""",
                value=oracledb.Timestamp(2002, 12, 14, 0, 0, 10, 250000))
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[5]])

    def test_2601_bind_null(self):
        "2601 - test binding in a null"
        self.cursor.setinputsizes(value=oracledb.DB_TYPE_TIMESTAMP)
        self.cursor.execute("""
                select * from TestTimestamps
                where TimestampCol = :value""",
                value=None)
        self.assertEqual(self.cursor.fetchall(), [])

    def test_2602_bind_out_set_input_sizes(self):
        "2602 - test binding out with set input sizes defined"
        bind_vars = self.cursor.setinputsizes(value=oracledb.DB_TYPE_TIMESTAMP)
        self.cursor.execute("""
                begin
                    :value := to_timestamp('20021209', 'YYYYMMDD');
                end;""")
        self.assertEqual(bind_vars["value"].getvalue(),
                         oracledb.Timestamp(2002, 12, 9))

    def test_2603_bind_in_out_set_input_sizes(self):
        "2603 - test binding in/out with set input sizes defined"
        bind_vars = self.cursor.setinputsizes(value=oracledb.DB_TYPE_TIMESTAMP)
        self.cursor.execute("""
                begin
                    :value := :value + 5.25;
                end;""",
                value = oracledb.Timestamp(2002, 12, 12, 10, 0, 0))
        self.assertEqual(bind_vars["value"].getvalue(),
                         oracledb.Timestamp(2002, 12, 17, 16, 0, 0))

    def test_2604_bind_out_var(self):
        "2604 - test binding out with cursor.var() method"
        var = self.cursor.var(oracledb.DB_TYPE_TIMESTAMP)
        self.cursor.execute("""
                begin
                    :value := to_date('20021231 12:31:00',
                        'YYYYMMDD HH24:MI:SS');
                end;""",
                value=var)
        self.assertEqual(var.getvalue(),
                         oracledb.Timestamp(2002, 12, 31, 12, 31, 0))

    def test_2605_bind_in_out_var_direct_set(self):
        "2605 - test binding in/out with cursor.var() method"
        var = self.cursor.var(oracledb.DB_TYPE_TIMESTAMP)
        var.setvalue(0, oracledb.Timestamp(2002, 12, 9, 6, 0, 0))
        self.cursor.execute("""
                begin
                    :value := :value + 5.25;
                end;""",
                value = var)
        self.assertEqual(var.getvalue(),
                         oracledb.Timestamp(2002, 12, 14, 12, 0, 0))

    def test_2606_cursor_description(self):
        "2606 - test cursor description is accurate"
        self.cursor.execute("select * from TestTimestamps")
        expected_value = [
            ('INTCOL', oracledb.DB_TYPE_NUMBER, 10, None, 9, 0, False),
            ('TIMESTAMPCOL', oracledb.DB_TYPE_TIMESTAMP, 23, None, 0, 6,
                    False),
            ('NULLABLECOL', oracledb.DB_TYPE_TIMESTAMP, 23, None, 0, 6, True)
        ]
        self.assertEqual(self.cursor.description, expected_value)

    def test_2607_fetchall(self):
        "2607 - test that fetching all of the data returns the correct results"
        self.cursor.execute("select * From TestTimestamps order by IntCol")
        self.assertEqual(self.cursor.fetchall(), self.raw_data)
        self.assertEqual(self.cursor.fetchall(), [])

    def test_2608_fetchmany(self):
        "2608 - test that fetching data in chunks returns the correct results"
        self.cursor.execute("select * From TestTimestamps order by IntCol")
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[0:3])
        self.assertEqual(self.cursor.fetchmany(2), self.raw_data[3:5])
        self.assertEqual(self.cursor.fetchmany(4), self.raw_data[5:9])
        self.assertEqual(self.cursor.fetchmany(3), self.raw_data[9:])
        self.assertEqual(self.cursor.fetchmany(3), [])

    def test_2609_fetchone(self):
        "2609 - test that fetching a single row returns the correct results"
        self.cursor.execute("""
                select *
                from TestTimestamps
                where IntCol in (3, 4)
                order by IntCol""")
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[3])
        self.assertEqual(self.cursor.fetchone(), self.data_by_key[4])
        self.assertEqual(self.cursor.fetchone(), None)

    def test_2610_bind_timestamp_with_zero_fseconds(self):
        "2610 - test binding a timestamp with zero fractional seconds"
        self.cursor.setinputsizes(value=oracledb.DB_TYPE_TIMESTAMP)
        self.cursor.execute("""
                select *
                from TestTimestamps
                where trunc(TimestampCol) = :value""",
                value=oracledb.Timestamp(2002, 12, 14))
        self.assertEqual(self.cursor.fetchall(), [self.data_by_key[5]])

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# protocol.pyx
#
# Cython file defining the protocol used by the client when communicating with
# the database (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

cdef class Protocol:

    cdef:
        uint8_t _seq_num
        object _socket
        Capabilities _caps
        ReadBuffer _read_buf
        WriteBuffer _write_buf
        object _request_lock
        bint _in_connect
        bint _txn_in_progress
        bint _break_in_progress

    def __init__(self):
        self._caps = Capabilities()
        # mark protocol to indicate that connect is in progress; this prevents
        # the normal break/reset mechanism from firing, which is unnecessary
        # since the connection is going to be immediately closed anyway!
        self._in_connect = True
        self._request_lock = threading.RLock()

    cdef int _break_external(self) except -1:
        """
        Method for sending a break to the server from an external request. A
        separate write buffer is used in order to avoid a potential conflict
        with any in progress writes.
        """
        cdef WriteBuffer buf
        if not self._break_in_progress:
            self._break_in_progress = True
            if self._caps.supports_oob:
                if DEBUG_PACKETS:
                    now = datetime.datetime.now()
                    print(now.isoformat(sep=" ", timespec="milliseconds"),
                          f"[socket: {self._socket.fileno()}]",
                          "Sending out of band break")
                    print()
                self._socket.send(b"!", socket.MSG_OOB)
            else:
                buf = WriteBuffer(self._socket, TNS_SDU, self._caps)
                self._send_marker(buf, TNS_MARKER_TYPE_INTERRUPT)

    cdef int _close(self, ThinConnImpl conn_impl) except -1:
        """
        Closes the connection. If a transaction is in progress it will be
        rolled back. DRCP sessions will be released. For standalone
        connections, the session will be logged off. For pooled connections,
        the connection will be returned to the pool for subsequent use.
        """
        cdef:
            Message message
            uint32_t release_mode = DRCP_DEAUTHENTICATE \
                    if conn_impl._pool is None else 0

        with self._request_lock:

            # if a read failed on the socket earlier, clear the socket
            if self._read_buf._socket is None:
                self._socket = None

            # if the session was marked as needing to be closed, force it
            # closed immediately (unless it was already closed)
            if self._read_buf._session_needs_to_be_closed \
                    and self._socket is not None:
                self._force_close()

            # rollback any open transaction and release the DRCP session, if
            # applicable
            if self._socket is not None:
                if self._txn_in_progress:
                    message = conn_impl._create_message(RollbackMessage)
                    self._process_message(message)
                if conn_impl._drcp_enabled:
                    self._release_drcp_session(conn_impl, release_mode)
                    conn_impl._drcp_establish_session = True

            # if the connection is part of a pool, return it to the pool
            if conn_impl._pool is not None:
                return conn_impl._pool._return_connection(conn_impl)

            # otherwise, destroy the database object type cache, send the
            # logoff message and final close packet
            if conn_impl._dbobject_type_cache_num > 0:
                remove_dbobject_type_cache(conn_impl._dbobject_type_cache_num)
                conn_impl._dbobject_type_cache_num = 0
            if self._socket is not None:
                if not conn_impl._drcp_enabled:
                    message = conn_impl._create_message(LogoffMessage)
                    self._process_message(message)
                self._final_close(self._write_buf)

    cdef int _connect_phase_one(self,
                                ThinConnImpl conn_impl,
                                ConnectParamsImpl params,
                                Description description,
                                Address address,
                                str connect_string) except -1:
        """
        Method for performing the required steps for establishing a connection
        within the scope of a retry. If the listener refuses the connection, a
        retry will be performed, if retry_count is set.
        """
        cdef:
            ConnectMessage connect_message = None
            uint8_t packet_type, packet_flags = 0
            object ssl_context, connect_info
            ConnectParamsImpl temp_params
            str host, redirect_data
            Address temp_address
            int port, pos

        # store whether OOB processing is possible or not
        self._caps.supports_oob = not params.disable_oob \
                and sys.platform != "win32"

        # establish initial TCP connection and get initial connect string
        host = address.host
        port = address.port
        self._connect_tcp(params, description, address, host, port)

        # send connect message and process response; this may request the
        # message to be resent multiple times; if a redirect packet is
        # detected, a new TCP connection is established first
        while True:

            # create connection message, if needed
            if connect_message is None:
                connect_message = conn_impl._create_message(ConnectMessage)
                connect_message.host = host
                connect_message.port = port
                connect_message.description = description
                connect_message.connect_string_bytes = connect_string.encode()
                connect_message.connect_string_len = \
                        <uint16_t> len(connect_message.connect_string_bytes)
                connect_message.packet_flags = packet_flags

            # process connection message
            self._process_message(connect_message)
            if connect_message.redirect_data is not None:
                redirect_data = connect_message.redirect_data
                pos = redirect_data.find('\x00')
                if pos < 0:
                    errors._raise_err(errors.ERR_INVALID_REDIRECT_DATA,
                                      data=redirect_data)
                temp_params = ConnectParamsImpl()
                temp_params._parse_connect_string(redirect_data[:pos])
                temp_address = temp_params._get_addresses()[0]
                host = temp_address.host
                port = temp_address.port
                connect_string = redirect_data[pos + 1:]
                self._connect_tcp(params, description, address, host, port)
                connect_message = None
                packet_flags = TNS_PACKET_FLAG_REDIRECT
            elif connect_message.packet_type == TNS_PACKET_TYPE_ACCEPT:
                break

            # for TCPS connections, OOB processing is not supported; if the
            # packet flags indicate that TLS renegotiation is required, this is
            # performed now
            if address.protocol == "tcps":
                self._caps.supports_oob = False
                if connect_message.packet_flags & TNS_PACKET_FLAG_TLS_RENEG:
                    ssl_context = self._socket.context
                    sock = socket.socket(fileno=self._socket.detach())
                    sock = perform_tls_negotiation(sock, ssl_context,
                                                   description, address)
                    self._set_socket(sock)

    cdef int _connect_phase_two(self, ThinConnImpl conn_impl,
                                Description description,
                                ConnectParamsImpl params) except -1:
        """"
        Method for perfoming the required steps for establishing a connection
        oustide the scope of a retry. If any of the steps in this method fail,
        an exception will be raised.
        """
        cdef:
            AuthMessage auth_message
        # check if the protocol version supported by the database is high
        # enough; if not, reject the connection immediately
        if self._caps.protocol_version < TNS_VERSION_MIN_ACCEPTED:
            errors._raise_err(errors.ERR_SERVER_VERSION_NOT_SUPPORTED)

        # if we can use OOB, send an urgent message now followed by a reset
        # marker to see if the server understands it
        if self._caps.supports_oob \
                and self._caps.protocol_version >= TNS_VERSION_MIN_OOB_CHECK:
            self._socket.send(b"!", socket.MSG_OOB)
            self._send_marker(self._write_buf, TNS_MARKER_TYPE_RESET)

        # send protocol and data types messages and process responses
        self._process_message(conn_impl._create_message(ProtocolMessage))
        self._process_message(conn_impl._create_message(DataTypesMessage))

        # send authorization packet twice, the first time to get the session
        # key and the second time to return the response to the challenge
        auth_message = conn_impl._create_message(AuthMessage)
        auth_message._set_params(params, description)
        self._process_message(auth_message)
        if auth_message.resend:
            self._process_message(auth_message)

        # mark protocol to indicate that connect is no longer in progress; this
        # allows the normal break/reset mechanism to fire
        self._in_connect = False

    cdef int _connect_tcp(self, ConnectParamsImpl params,
                          Description description, Address address, str host,
                          int port) except -1:
        """
        Creates a socket on which to communicate using the provided parameters.
        If a proxy is configured, a connection to the proxy is established and
        the target host and port is forwarded to the proxy.
        """
        cdef:
            bint use_proxy = (address.https_proxy is not None)
            double timeout = description.tcp_connect_timeout
            bint use_tcps = (address.protocol == "tcps")
            object connect_info, sock, data, reply, m

        # establish connection to appropriate host/port
        if use_proxy:
            if not use_tcps:
                errors._raise_err(errors.ERR_HTTPS_PROXY_REQUIRES_TCPS)
            connect_info = (address.https_proxy, address.https_proxy_port)
        else:
            connect_info = (host, port)
            if not use_tcps and (params._token is not None
                    or params.access_token_callback is not None):
                errors._raise_err(errors.ERR_ACCESS_TOKEN_REQUIRES_TCPS)
        sock = socket.create_connection(connect_info, timeout)

        # complete connection through proxy, if applicable
        if use_proxy:
            data = f"CONNECT {host}:{port} HTTP/1.0\r\n\r\n"
            sock.send(data.encode())
            reply = sock.recv(1024)
            m = re.search('HTTP/1.[01]\\s+(\\d+)\\s+', reply.decode())
            if m is None or m.groups()[0] != '200':
                errors._raise_err(errors.ERR_PROXY_FAILURE,
                                  response=reply.decode())

        # set socket options
        if description.expire_time > 0:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            if hasattr(socket, "TCP_KEEPIDLE") \
                    and hasattr(socket, "TCP_KEEPINTVL") \
                    and hasattr(socket, "TCP_KEEPCNT"):
                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE,
                                description.expire_time * 60)
                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 6)
                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 10)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        sock.settimeout(None)

        # establish TLS connection, if applicable
        if use_tcps:
            sock = get_ssl_socket(sock, params, description, address)

        # save final socket object
        self._set_socket(sock)

    cdef int _final_close(self, WriteBuffer buf) except -1:
        """
        Send the final close packet to the server and close the socket.
        """
        buf.start_request(TNS_PACKET_TYPE_DATA, 0, TNS_DATA_FLAGS_EOF)
        buf.end_request()
        self._socket.shutdown(socket.SHUT_RDWR)
        self._socket.close()
        self._socket = None

    cdef int _force_close(self) except -1:
        """
        Forces the connection closed. This is used when an unrecoverable error
        has taken place.
        """
        sock = self._socket
        if DEBUG_PACKETS:
            now = datetime.datetime.now()
            print(now.isoformat(sep=" ", timespec="milliseconds"),
                  f"[socket: {sock.fileno()}]", "force closing connection")
            print()
        self._socket = None
        self._read_buf._socket = None
        self._write_buf._socket = None
        sock.shutdown(socket.SHUT_RDWR)
        sock.close()

    cdef int _process_message(self, Message message) except -1:
        try:
            message.send(self._write_buf)
            self._receive_packet(message)
            message.process(self._read_buf)
        except socket.timeout:
            try:
                self._break_external()
                self._receive_packet(message)
                self._break_in_progress = False
                timeout = int(self._socket.gettimeout() * 1000)
                errors._raise_err(errors.ERR_CALL_TIMEOUT_EXCEEDED,
                                  timeout=timeout)
            except socket.timeout:
                self._force_close()
                errors._raise_err(errors.ERR_CONNECTION_CLOSED,
                                  "socket timed out while recovering from " \
                                  "previous socket timeout")
            raise
        except:
            if not self._in_connect \
                    and self._write_buf._packet_sent \
                    and self._read_buf._socket is not None:
                self._send_marker(self._write_buf, TNS_MARKER_TYPE_BREAK)
                self._reset(message)
            raise
        if message.flush_out_binds:
            self._write_buf.start_request(TNS_PACKET_TYPE_DATA)
            self._write_buf.write_uint8(TNS_MSG_TYPE_FLUSH_OUT_BINDS)
            self._write_buf.end_request()
            self._reset(message)
            message.process(self._read_buf)
        if self._break_in_progress:
            try:
                if self._caps.supports_oob:
                    self._send_marker(self._write_buf,
                                      TNS_MARKER_TYPE_INTERRUPT)
                self._receive_packet(message)
            except socket.timeout:
                errors._raise_err(errors.ERR_CONNECTION_CLOSED,
                                  "socket timed out while awaiting break " \
                                  "response from server")
            message.process(self._read_buf)
            self._break_in_progress = False
        self._txn_in_progress = message.call_status & TNS_TXN_IN_PROGRESS
        if message.error_occurred:
            if message.error_info.num == TNS_ERR_VAR_NOT_IN_SELECT_LIST:
                message.error_occurred = False
                return self._process_message(message)
            error = errors._Error(message.error_info.message,
                                  code=message.error_info.num,
                                  offset=message.error_info.pos)

            # if a connection has received dead connection error then it is no
            # longer usable
            if error.is_session_dead:
                self._force_close()
            exc_type = get_exception_class(message.error_info.num)
            raise exc_type(error)

    cdef int _process_single_message(self, Message message) except -1:
        """
        Process a single message within a request.
        """
        with self._request_lock:
            self._process_message(message)
            if message.resend:
                self._process_message(message)

    cdef int _receive_packet(self, Message message) except -1:
        cdef ReadBuffer buf = self._read_buf
        buf.receive_packet(&message.packet_type, &message.packet_flags)
        if message.packet_type == TNS_PACKET_TYPE_MARKER:
            self._reset(message)
        elif message.packet_type == TNS_PACKET_TYPE_REFUSE:
            self._write_buf._packet_sent = False
            buf.skip_raw_bytes(3)
            message.error_info.message = buf.read_str(TNS_CS_IMPLICIT)

    cdef int _release_drcp_session(self, ThinConnImpl conn_impl,
                                   uint32_t release_mode) except -1:
        """
        Release the session back to DRCP. Standalone sessions are marked for
        deauthentication.
        """
        cdef SessionReleaseMessage message
        message = conn_impl._create_message(SessionReleaseMessage)
        message.release_mode = release_mode
        message.send(self._write_buf)

    cdef int _reset(self, Message message) except -1:
        cdef uint8_t marker_type

        # send reset marker
        self._send_marker(self._write_buf, TNS_MARKER_TYPE_RESET)

        # read and discard all packets until a marker packet is received
        while True:
            if message.packet_type == TNS_PACKET_TYPE_MARKER:
                self._read_buf.skip_raw_bytes(2)
                self._read_buf.read_ub1(&marker_type)
                if marker_type == TNS_MARKER_TYPE_RESET:
                    break
            self._read_buf.receive_packet(&message.packet_type,
                                          &message.packet_flags)

        # read error packet; first skip as many marker packets as may be sent
        # by the server; if the server doesn't handle out-of-band breaks
        # properly, some quit immediately and others send multiple reset
        # markers (this addresses both situations without resulting in strange
        # errors)
        while message.packet_type == TNS_PACKET_TYPE_MARKER:
            self._read_buf.receive_packet(&message.packet_type,
                                          &message.packet_flags)
        self._break_in_progress = False

    cdef int _send_marker(self, WriteBuffer buf, uint8_t marker_type):
        """
        Sends a marker of the specified type to the server.
        Internal method for sending a break to the server.
        """
        buf.start_request(TNS_PACKET_TYPE_MARKER)
        buf.write_uint8(1)
        buf.write_uint8(0)
        buf.write_uint8(marker_type)
        buf.end_request()

    cdef int _set_socket(self, sock):
         self._socket = sock
         self._read_buf = ReadBuffer(sock, TNS_SDU, self._caps)
         self._write_buf = WriteBuffer(sock, TNS_SDU, self._caps)



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
3100 - Module for testing boolean variables
"""

import unittest

import oracledb
import test_env

@unittest.skipUnless(test_env.get_client_version() >= (12, 1),
                     "unsupported client")
class TestCase(test_env.BaseTestCase):

    def __test_bind_value_as_boolean(self, value):
        expected_result = str(bool(value)).upper()
        var = self.cursor.var(bool)
        var.setvalue(0, value)
        result = self.cursor.callfunc("pkg_TestBooleans.GetStringRep", str,
                                      (var,))
        self.assertEqual(result, expected_result)

    def test_3100_bind_false(self):
        "3100 - test binding in a False value"
        result = self.cursor.callfunc("pkg_TestBooleans.GetStringRep", str,
                                      (False,))
        self.assertEqual(result, "FALSE")

    def test_3101_bind_float_as_boolean(self):
        "3101 - test binding in a float as a boolean"
        self.__test_bind_value_as_boolean(0.0)
        self.__test_bind_value_as_boolean(1.0)

    def test_3102_bind_integer_as_boolean(self):
        "3102 - test binding in an integer as a boolean"
        self.__test_bind_value_as_boolean(0)
        self.__test_bind_value_as_boolean(1)

    def test_3103_bind_null(self):
        "3103 - test binding in a null value"
        self.cursor.setinputsizes(None, bool)
        result = self.cursor.callfunc("pkg_TestBooleans.GetStringRep", str,
                                      (None,))
        self.assertEqual(result, "NULL")

    def test_3104_bind_out_false(self):
        "3104 - test binding out a boolean value (False)"
        result = self.cursor.callfunc("pkg_TestBooleans.IsLessThan10",
                                      oracledb.DB_TYPE_BOOLEAN, (15,))
        self.assertEqual(result, False)

    def test_3105_bind_out_true(self):
        "3105 - test binding out a boolean value (True)"
        result = self.cursor.callfunc("pkg_TestBooleans.IsLessThan10", bool,
                                      (5,))
        self.assertEqual(result, True)

    def test_3106_bind_string_as_boolean(self):
        "3106 - test binding in a string as a boolean"
        self.__test_bind_value_as_boolean("")
        self.__test_bind_value_as_boolean("0")

    def test_3107_bind_true(self):
        "3107 - test binding in a True value"
        result = self.cursor.callfunc("pkg_TestBooleans.GetStringRep", str,
                                      (True,))
        self.assertEqual(result, "TRUE")

    def test_3108_bind_out_null(self):
        "3108 - test binding out a boolean value (None)"
        result = self.cursor.callfunc("pkg_TestBooleans.TestOutValueNull",
                                      bool)
        self.assertEqual(result, None)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
3900 - Module for testing the cursor execute() method
"""

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):

    def test_3900_execute_no_args(self):
        "3900 - test executing a statement without any arguments"
        result = self.cursor.execute("begin null; end;")
        self.assertEqual(result, None)

    def test_3901_execute_no_statement_with_args(self):
        "3901 - test executing a None statement with bind variables"
        cursor = self.connection.cursor()
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2001:",
                               cursor.execute, None, x=5)

    def test_3902_execute_empty_keyword_args(self):
        "3902 - test executing a statement with args and empty keyword args"
        simple_var = self.cursor.var(oracledb.NUMBER)
        args = [simple_var]
        kwargs = {}
        result = self.cursor.execute("begin :1 := 25; end;", args, **kwargs)
        self.assertEqual(result, None)
        self.assertEqual(simple_var.getvalue(), 25)

    def test_3903_execute_keyword_args(self):
        "3903 - test executing a statement with keyword arguments"
        simple_var = self.cursor.var(oracledb.NUMBER)
        result = self.cursor.execute("begin :value := 5; end;",
                                     value=simple_var)
        self.assertEqual(result, None)
        self.assertEqual(simple_var.getvalue(), 5)

    def test_3904_execute_dictionary_arg(self):
        "3904 - test executing a statement with a dictionary argument"
        simple_var = self.cursor.var(oracledb.NUMBER)
        dict_arg = dict(value=simple_var)
        result = self.cursor.execute("begin :value := 10; end;", dict_arg)
        self.assertEqual(result, None)
        self.assertEqual(simple_var.getvalue(), 10)

    def test_3905_execute_multiple_arg_types(self):
        "3905 - test executing a statement with both a dict and keyword args"
        simple_var = self.cursor.var(oracledb.NUMBER)
        dict_arg = dict(value=simple_var)
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2005:",
                               self.cursor.execute,
                               "begin :value := 15; end;", dict_arg,
                               value=simple_var)

    def test_3906_execute_and_modify_array_size(self):
        "3906 - test executing a statement and then changing the array size"
        self.cursor.execute("select IntCol from TestNumbers")
        self.cursor.arraysize = 20
        self.assertEqual(len(self.cursor.fetchall()), 10)

    def test_3907_bad_execute(self):
        "3907 - test that subsequent executes succeed after bad execute"
        sql = "begin raise_application_error(-20000, 'this); end;"
        self.assertRaisesRegex(oracledb.DatabaseError,
                               "^ORA-06550:|^ORA-01756:", self.cursor.execute,
                               sql)
        self.cursor.execute("begin null; end;")

    def test_3908_fetch_after_bad_execute(self):
        "3908 - test that subsequent fetches fail after bad execute"
        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-00904:",
                               self.cursor.execute, "select y from dual")
        self.assertRaisesRegex(oracledb.InterfaceError, "^DPY-1003:",
                               self.cursor.fetchall)

    def test_3909_execute_bind_names_with_incorrect_bind(self):
        "3909 - test executing a statement with an incorrect named bind"
        statement = "select * from TestStrings where IntCol = :value"
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-4008:|^ORA-01036:",
                               self.cursor.execute, statement, value2=3)

    def test_3910_execute_with_named_binds(self):
        "3910 - test executing a statement with named binds"
        statement = "select * from TestNumbers where IntCol = :value1 " + \
                    "and LongIntCol = :value2"
        result = self.cursor.execute(statement, value1=1, value2=38)
        self.assertEqual(len(result.fetchall()), 1)

    def test_3911_execute_bind_position_with_incorrect_bind(self):
        "3911 - test executing a statement with an incorrect positional bind"
        statement = "select * from TestNumbers where IntCol = :value " + \
                    "and LongIntCol = :value2"
        self.assertRaisesRegex(oracledb.DatabaseError,
                               "^DPY-4009:|^ORA-01008:",
                               self.cursor.execute, statement, [3])

    def test_3912_execute_with_positional_binds(self):
        "3912 - test executing a statement with positional binds"
        statement = "select * from TestNumbers where IntCol = :value " + \
                    "and LongIntCol = :value2"
        result = self.cursor.execute(statement, [1,38])
        self.assertEqual(len(result.fetchall()), 1)

    def test_3913_execute_with_rebinding_bind_name(self):
        "3913 - test executing a statement after rebinding a named bind"
        statement = "begin :value := :value2 + 5; end;"
        simple_var = self.cursor.var(oracledb.NUMBER)
        simple_var2 = self.cursor.var(oracledb.NUMBER)
        simple_var2.setvalue(0, 5)
        result = self.cursor.execute(statement, value=simple_var,
                                     value2=simple_var2)
        self.assertEqual(result, None)
        self.assertEqual(simple_var.getvalue(), 10)

        simple_var = self.cursor.var(oracledb.NATIVE_FLOAT)
        simple_var2 = self.cursor.var(oracledb.NATIVE_FLOAT)
        simple_var2.setvalue(0, 10)
        result = self.cursor.execute(statement, value=simple_var,
                                     value2=simple_var2)
        self.assertEqual(result, None)
        self.assertEqual(simple_var.getvalue(), 15)

    def test_3914_bind_by_name_with_duplicates(self):
        "3914 - test executing a PL/SQL statement with duplicate binds"
        statement = "begin :value := :value + 5; end;"
        simple_var = self.cursor.var(oracledb.NUMBER)
        simple_var.setvalue(0, 5)
        result = self.cursor.execute(statement, value=simple_var)
        self.assertEqual(result, None)
        self.assertEqual(simple_var.getvalue(), 10)

    def test_3915_positional_bind_with_duplicates(self):
        "3915 - test executing a PL/SQL statement with duplicate binds"
        statement = "begin :value := :value + 5; end;"
        simple_var = self.cursor.var(oracledb.NUMBER)
        simple_var.setvalue(0, 5)
        self.cursor.execute(statement, [simple_var])
        self.assertEqual(simple_var.getvalue(), 10)

    def test_3916_execute_with_incorrect_bind_values(self):
        "3916 - test executing a statement with an incorrect number of binds"
        statement = "begin :value := :value2 + 5; end;"
        var = self.cursor.var(oracledb.NUMBER)
        var.setvalue(0, 5)
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-4010:|^ORA-01008:",
                               self.cursor.execute, statement)
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-4010:|^ORA-01008:",
                               self.cursor.execute, statement, value=var)
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-4008:|^ORA-01036:",
                               self.cursor.execute, statement, value=var,
                               value2=var, value3=var)

    def test_3917_change_in_size_on_successive_bind(self):
        "3917 - change in size on subsequent binds does not use optimised path"
        self.cursor.execute("truncate table TestTempTable")
        data = [
            (1, "Test String #1"),
            (2, "ABC" * 100)
        ]
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        for row in data:
            self.cursor.execute(sql, row)
        self.connection.commit()
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(self.cursor.fetchall(), data)

    def test_3918_dml_can_use_optimised_path(self):
        "3918 - test that dml can use optimised path"
        data_to_insert = [
            (1, "Test String #1"),
            (2, "Test String #2"),
            (3, "Test String #3")
        ]
        self.cursor.execute("truncate table TestTempTable")
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        for row in data_to_insert:
            with self.connection.cursor() as cursor:
                cursor.execute(sql, row)
        self.connection.commit()
        self.cursor.execute("""
                select IntCol, StringCol1
                from TestTempTable
                order by IntCol""")
        self.assertEqual(self.cursor.fetchall(), data_to_insert)

    def test_3919_execute_with_invalid_parameters(self):
        "3919 - test calling execute() with invalid parameters"
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2003:",
                               self.cursor.execute, sql,
                               "These are not valid parameters")

    def test_3920_execute_with_mixed_binds(self):
        "3920 - test calling execute() with mixed binds"
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.setinputsizes(None, None, str)
        data = dict(val1=1, val2="Test String 1")
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2006:",
                               self.cursor.execute, """
                               insert into TestTempTable (IntCol, StringCol1)
                               values (:1, :2)
                               returning StringCol1 into :out_var""", data)

    def test_3921_bind_by_name_with_double_quotes(self):
        "3921 - test binding by name with double quotes"
        sql = 'select :"_value1" + :"VaLue_2" + :"3VALUE" from dual'
        data = {'"_value1"': 1, '"VaLue_2"': 2, '"3VALUE"': 3}
        self.cursor.execute(sql, data)
        result, = self.cursor.fetchone()
        self.assertEqual(result, 6)

    def test_3922_resize_buffer(self):
        "3922 - test executing a statement with different input buffer sizes"
        sql = """
                insert into TestTempTable (IntCol, StringCol1, StringCol2)
                values (:int_col, :str_val1, :str_val2) returning IntCol
                into :ret_data"""
        values1 = {"int_col" : 1, "str_val1": '{"a", "b"}', "str_val2" : None}
        values2 = {"int_col" : 2, "str_val1": None, "str_val2" : '{"a", "b"}'}
        values3 = {"int_col" : 3, "str_val1": '{"a"}', "str_val2" : None}

        self.cursor.execute("truncate table TestTempTable")
        ret_bind = self.cursor.var(oracledb.DB_TYPE_VARCHAR, arraysize=1)
        self.cursor.setinputsizes(ret_data=ret_bind)
        self.cursor.execute(sql, values1)
        self.assertEqual(ret_bind.values, [['1']])

        ret_bind = self.cursor.var(oracledb.DB_TYPE_VARCHAR, arraysize=1)
        self.cursor.setinputsizes(ret_data=ret_bind)
        self.cursor.execute(sql, values2)
        self.assertEqual(ret_bind.values, [['2']])

        ret_bind = self.cursor.var(oracledb.DB_TYPE_VARCHAR, arraysize=1)
        self.cursor.setinputsizes(ret_data=ret_bind)
        self.cursor.execute(sql, values3)
        self.assertEqual(ret_bind.values, [['3']])

    def test_3923_rowfactory_callable(self):
        "3923 - test using rowfactory"
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.execute("""
                insert into TestTempTable (IntCol, StringCol1)
                values (1, 'Test 1')""")
        self.connection.commit()
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        column_names = [col[0] for col in self.cursor.description]
        self.cursor.rowfactory = lambda *row: dict(zip(column_names, row))
        self.assertEqual(self.cursor.fetchall(),
                         [{'INTCOL': 1, 'STRINGCOL1': 'Test 1'}])

    def test_3924_rowfactory_execute_same_sql(self):
        "3924 - test executing same query after setting rowfactory"
        self.cursor.execute("truncate table TestTempTable")
        data = [
            (1, 'Test 1'),
            (2, 'Test 2')
        ]
        self.cursor.executemany("""
                insert into TestTempTable (IntCol, StringCol1)
                values (:1, :2)""", data)
        self.connection.commit()
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        column_names = [col[0] for col in self.cursor.description]
        self.cursor.rowfactory = lambda *row: dict(zip(column_names, row))
        results_1 = self.cursor.fetchall()
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        results_2 = self.cursor.fetchall()
        self.assertEqual(results_1, results_2)

    def test_3925_rowfactory_execute_different_sql(self):
        "3925 - test executing different query after setting rowfactory"
        self.cursor.execute("truncate table TestTempTable")
        data = [
            (1, 'Test 1'),
            (2, 'Test 2')
        ]
        self.cursor.executemany("""
                insert into TestTempTable (IntCol, StringCol1)
                values (:1, :2)""", data)
        self.connection.commit()
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        column_names = [col[0] for col in self.cursor.description]
        self.cursor.rowfactory = lambda *row: dict(zip(column_names, row))
        self.cursor.execute("""
                select IntCol, StringCol from TestSTrings
                where IntCol between 1 and 3 order by IntCol""")
        expected_data = [
            (1, 'String 1'),
            (2, 'String 2'),
            (3, 'String 3')
        ]
        self.assertEqual(self.cursor.fetchall(), expected_data)

    def test_3926_rowfactory_on_refcursor(self):
        "3926 - test setting rowfactory on a REF cursor"
        with self.connection.cursor() as cursor:
            ref_cursor = cursor.callfunc("pkg_TestRefCursors.TestReturnCursor",
                                         oracledb.DB_TYPE_CURSOR, [2])
            column_names = [col[0] for col in ref_cursor.description]
            ref_cursor.rowfactory = lambda *row: dict(zip(column_names, row))
            self.assertEqual(ref_cursor.fetchall(),
                             [{'INTCOL': 1, 'STRINGCOL': 'String 1'},
                              {'INTCOL': 2, 'STRINGCOL': 'String 2'}])

    def test_3927_subclassed_string(self):
        "3927 - test using a subclassed string as bind parameter keys"
        class my_str(str):
            pass
        self.cursor.execute("truncate table TestTempTable")
        keys = {my_str("str_val"): oracledb.DB_TYPE_VARCHAR}
        self.cursor.setinputsizes(**keys)
        values = {
            my_str("int_val"): 3927,
            my_str("str_val"): "3927 - String Value"
        }
        self.cursor.execute("""
                insert into TestTempTable (IntCol, StringCol1)
                values (:int_val, :str_val)""", values)
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(self.cursor.fetchall(),
                         [(3927, "3927 - String Value")])

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# pool_params.py
#
# Contains the PoolParams class used for managing the parameters required to
# create a connection pool.
#
# *** NOTICE *** This file is generated from a template and should not be
# modified directly. See build_from_template.py in the utils subdirectory for
# more information.
#------------------------------------------------------------------------------

from typing import Callable, Type, Union

import oracledb

from . import base_impl, constants, errors, utils
from . import connection as connection_module
from .connect_params import ConnectParams

class PoolParams(ConnectParams):
    """
    Contains all parameters used for creating a connection pool.
    """
    __module__ = oracledb.__name__
    __slots__ = ["_impl"]
    _impl_class = base_impl.PoolParamsImpl

    @utils.params_initer
    def __init__(self, *,
                 min: int=1,
                 max: int=2,
                 increment: int=1,
                 connectiontype: Type["oracledb.Connection"]=None,
                 getmode: int=oracledb.POOL_GETMODE_WAIT,
                 homogeneous: bool=True,
                 timeout: int=0,
                 wait_timeout: int=0,
                 max_lifetime_session: int=0,
                 session_callback: Callable=None,
                 max_sessions_per_shard: int=0,
                 soda_metadata_cache: bool=False,
                 ping_interval: int=60,
                 user: str=None,
                 proxy_user: str=None,
                 password: str=None,
                 newpassword: str=None,
                 wallet_password: str=None,
                 access_token: Union[str, tuple, Callable]=None,
                 host: str=None,
                 port: int=1521,
                 protocol: str="tcp",
                 https_proxy: str=None,
                 https_proxy_port: int=0,
                 service_name: str=None,
                 sid: str=None,
                 server_type: str=None,
                 cclass: str=None,
                 purity: int=oracledb.PURITY_DEFAULT,
                 expire_time: int=0,
                 retry_count: int=0,
                 retry_delay: int=0,
                 tcp_connect_timeout: float=60.0,
                 ssl_server_dn_match: bool=True,
                 ssl_server_cert_dn: str=None,
                 wallet_location: str=None,
                 events: bool=False,
                 externalauth: bool=False,
                 mode: int=oracledb.AUTH_MODE_DEFAULT,
                 disable_oob: bool=False,
                 stmtcachesize: int=oracledb.defaults.stmtcachesize,
                 edition: str=None,
                 tag: str=None,
                 matchanytag: bool=False,
                 config_dir: str=oracledb.defaults.config_dir,
                 appcontext: list=None,
                 shardingkey: list=None,
                 supershardingkey: list=None,
                 debug_jdwp: str=None,
                 handle: int=0,
                 threaded: bool=True,
                 encoding: str=None,
                 nencoding: str=None,
                 waitTimeout: int=None,
                 maxLifetimeSession: int=None,
                 maxSessionsPerShard: int=None,
                 sessionCallback: Callable=None
                ):
        """
        All parameters are optional. A brief description of each parameter
        follows:

        - min: the minimum number of connections the pool should contain
          (default: 1)

        - max: the maximum number of connections the pool should contain
          (default: 2)

        - increment: the number of connections that should be added to the pool
          whenever a new connection needs to be created (default: 1)

        - connectiontype: the class of the connection that should be returned
          during calls to pool.acquire(). It must be Connection or a subclass
          of Connection (default: None)

        - getmode: how pool.acquire() will behave. One of the constants
          oracledb.POOL_GETMODE_WAIT, oracledb.POOL_GETMODE_NOWAIT,
          oracledb.POOL_GETMODE_FORCEGET, or oracledb.POOL_GETMODE_TIMEDWAIT
          (default: oracledb.POOL_GETMODE_WAIT)

        - homogeneous: a boolean indicating whether the connections are
          homogeneous (same user) or heterogeneous (multiple users) (default:
          True)

        - timeout: length of time (in seconds) that a connection may remain
          idle in the pool before it is terminated. If it is 0 then connections
          are never terminated (default: 0)

        - wait_timeout: length of time (in milliseconds) that a caller should
          wait when acquiring a connection from the pool with getmode set to
          oracledb.POOL_GETMODE_TIMEDWAIT (default: 0)

        - max_lifetime_session: length of time (in seconds) that connections
          can remain in the pool. If it is 0 then connections may remain in the
          pool indefinitely (default: 0)

        - session_callback: a callable that is invoked when a connection is
          returned from the pool for the first time, or when the connection tag
          differs from the one requested (default: None)

        - max_sessions_per_shard: the maximum number of connections that may be
          associated with a particular shard (default: 0)

        - soda_metadata_cache: boolean indicating whether or not the SODA
          metadata cache should be enabled (default: False)

        - ping_interval: length of time (in seconds) after which an unused
          connection in the pool will be a candidate for pinging when
          pool.acquire() is called. If the ping to the database indicates the
          connection is not alive a replacement connection will be returned by
          pool.acquire(). If ping_interval is a negative value the ping
          functionality will be disabled (default: 60)

        - user: the name of the user to connect to (default: None)

        - proxy_user: the name of the proxy user to connect to. If this value
          is not specified, it will be parsed out of user if user is in the
          form "user[proxy_user]" (default: None)

        - password: the password for the user (default: None)

        - newpassword: the new password for the user. The new password will
          take effect immediately upon a successful connection to the database
          (default: None)

        - wallet_password: the password to use to decrypt the wallet, if it is
          encrypted. This value is only used in thin mode (default: None)

        - access_token: expected to be a string or a 2-tuple or a callable. If
          it is a string, it specifies an Azure AD OAuth2 token used for Open
          Authorization (OAuth 2.0) token based authentication. If it is a
          2-tuple, it specifies the token and private key strings used for
          Oracle Cloud Infrastructure (OCI) Identity and Access Management
          (IAM) token based authentication. If it is a callable, it returns
          either a string or a 2-tuple used for OAuth 2.0 or OCI IAM token
          based authentication and is useful when the pool needs to expand and
          create new connections but the current authentication token has
          expired (default: None)

        - host: the name or IP address of the machine hosting the database or
          the database listener (default: None)

        - port: the port number on which the database listener is listening
          (default: 1521)

        - protocol: one of the strings "tcp" or "tcps" indicating whether to
          use unencrypted network traffic or encrypted network traffic (TLS)
          (default: "tcp")

        - https_proxy: the name or IP address of a proxy host to use for
          tunneling secure connections (default: None)

        - https_proxy_port: the port on which to communicate with the proxy
          host (default: 0)

        - service_name: the service name of the database (default: None)

        - sid: the system identifier (SID) of the database. Note using a
          service_name instead is recommended (default: None)

        - server_type: the type of server connection that should be
          established. If specified, it should be one of "dedicated", "shared"
          or "pooled" (default: None)

        - cclass: connection class to use for Database Resident Connection
          Pooling (DRCP) (default: None)

        - purity: purity to use for Database Resident Connection Pooling (DRCP)
          (default: oracledb.PURITY_DEFAULT)

        - expire_time: an integer indicating the number of minutes between the
          sending of keepalive probes. If this parameter is set to a value
          greater than zero it enables keepalive (default: 0)

        - retry_count: the number of times that a connection attempt should be
          retried before the attempt is terminated (default: 0)

        - retry_delay: the number of seconds to wait before making a new
          connection attempt (default: 0)

        - tcp_connect_timeout: a float indicating the maximum number of seconds
          to wait for establishing a connection to the database host (default:
          60.0)

        - ssl_server_dn_match: boolean indicating whether the server
          certificate distinguished name (DN) should be matched in addition to
          the regular certificate verification that is performed. Note that if
          the ssl_server_cert_dn parameter is not privided, host name matching
          is performed instead (default: True)

        - ssl_server_cert_dn: the distinguished name (DN) which should be
          matched with the server. This value is ignored if the
          ssl_server_dn_match parameter is not set to the value True. If
          specified this value is used for any verfication. Otherwise the
          hostname will be used. (default: None)

        - wallet_location: the directory where the wallet can be found. In thin
          mode this must be the directory containing the PEM-encoded wallet
          file ewallet.pem. In thick mode this must be the directory containing
          the file cwallet.sso (default: None)

        - events: boolean specifying whether events mode should be enabled.
          This value is only used in thick mode and is needed for continuous
          query notification and high availability event notifications
          (default: False)

        - externalauth: a boolean indicating whether to use external
          authentication (default: False)

        - mode: authorization mode to use. For example
          oracledb.AUTH_MODE_SYSDBA (default: oracledb.AUTH_MODE_DEFAULT)

        - disable_oob: boolean indicating whether out-of-band breaks should be
          disabled. This value is only used in thin mode. It has no effect on
          Windows which does not support this functionality (default: False)

        - stmtcachesize: identifies the initial size of the statement cache
          (default: oracledb.defaults.stmtcachesize)

        - edition: edition to use for the connection. This parameter cannot be
          used simultaneously with the cclass parameter (default: None)

        - tag: identifies the type of connection that should be returned from a
          pool. This value is only used in thick mode (default: None)

        - matchanytag: boolean specifying whether any tag can be used when
          acquiring a connection from the pool. This value is only used in
          thick mode. (default: False)

        - config_dir: directory in which the optional tnsnames.ora
          configuration file is located. This value is only used in thin mode.
          For thick mode use the config_dir parameter of init_oracle_client()
          (default: oracledb.defaults.config_dir)

        - appcontext: application context used by the connection. It should be
          a list of 3-tuples (namespace, name, value) and each entry in the
          tuple should be a string. This value is only used in thick mode
          (default: None)

        - shardingkey: a list of strings, numbers, bytes or dates that identify
          the database shard to connect to. This value is only used in thick
          mode (default: None)

        - supershardingkey: a list of strings, numbers, bytes or dates that
          identify the database shard to connect to. This value is only used in
          thick mode (default: None)

        - debug_jdwp: a string with the format "host=<host>;port=<port>" that
          specifies the host and port of the PL/SQL debugger. This value is
          only used in thin mode. For thick mode set the ORA_DEBUG_JDWP
          environment variable (default: None)

        - handle: an integer representing a pointer to a valid service context
          handle. This value is only used in thick mode. It should be used with
          extreme caution (default: 0)
        """
        pass

    def __repr__(self):
        return self.__class__.__qualname__ + "(" + \
               f"min={self.min!r}, " + \
               f"max={self.max!r}, " + \
               f"increment={self.increment!r}, " + \
               f"connectiontype={self.connectiontype!r}, " + \
               f"getmode={self.getmode!r}, " + \
               f"homogeneous={self.homogeneous!r}, " + \
               f"timeout={self.timeout!r}, " + \
               f"wait_timeout={self.wait_timeout!r}, " + \
               f"max_lifetime_session={self.max_lifetime_session!r}, " + \
               f"session_callback={self.session_callback!r}, " + \
               f"max_sessions_per_shard={self.max_sessions_per_shard!r}, " + \
               f"soda_metadata_cache={self.soda_metadata_cache!r}, " + \
               f"ping_interval={self.ping_interval!r}, " + \
               f"user={self.user!r}, " + \
               f"proxy_user={self.proxy_user!r}, " + \
               f"host={self.host!r}, " + \
               f"port={self.port!r}, " + \
               f"protocol={self.protocol!r}, " + \
               f"https_proxy={self.https_proxy!r}, " + \
               f"https_proxy_port={self.https_proxy_port!r}, " + \
               f"service_name={self.service_name!r}, " + \
               f"sid={self.sid!r}, " + \
               f"server_type={self.server_type!r}, " + \
               f"cclass={self.cclass!r}, " + \
               f"purity={self.purity!r}, " + \
               f"expire_time={self.expire_time!r}, " + \
               f"retry_count={self.retry_count!r}, " + \
               f"retry_delay={self.retry_delay!r}, " + \
               f"tcp_connect_timeout={self.tcp_connect_timeout!r}, " + \
               f"ssl_server_dn_match={self.ssl_server_dn_match!r}, " + \
               f"ssl_server_cert_dn={self.ssl_server_cert_dn!r}, " + \
               f"wallet_location={self.wallet_location!r}, " + \
               f"events={self.events!r}, " + \
               f"externalauth={self.externalauth!r}, " + \
               f"mode={self.mode!r}, " + \
               f"disable_oob={self.disable_oob!r}, " + \
               f"stmtcachesize={self.stmtcachesize!r}, " + \
               f"edition={self.edition!r}, " + \
               f"tag={self.tag!r}, " + \
               f"matchanytag={self.matchanytag!r}, " + \
               f"config_dir={self.config_dir!r}, " + \
               f"appcontext={self.appcontext!r}, " + \
               f"shardingkey={self.shardingkey!r}, " + \
               f"supershardingkey={self.supershardingkey!r}, " + \
               f"debug_jdwp={self.debug_jdwp!r}" + \
               ")"

    @property
    def connectiontype(self) -> Type["oracledb.Connection"]:
        """
        The class of the connection that should be returned during calls to
        pool.acquire(). It must be Connection or a subclass of Connection.
        """
        return self._impl.connectiontype

    @property
    def getmode(self) -> int:
        """
        How pool.acquire() will behave. One of the constants
        oracledb.POOL_GETMODE_WAIT, oracledb.POOL_GETMODE_NOWAIT,
        oracledb.POOL_GETMODE_FORCEGET, or oracledb.POOL_GETMODE_TIMEDWAIT.
        """
        return self._impl.getmode

    @property
    def homogeneous(self) -> bool:
        """
        A boolean indicating whether the connections are homogeneous (same
        user) or heterogeneous (multiple users).
        """
        return self._impl.homogeneous

    @property
    def increment(self) -> int:
        """
        The number of connections that should be added to the pool whenever a
        new connection needs to be created.
        """
        return self._impl.increment

    @property
    def max(self) -> int:
        """
        The maximum number of connections the pool should contain.
        """
        return self._impl.max

    @property
    def max_lifetime_session(self) -> int:
        """
        Length of time (in seconds) that connections can remain in the pool. If
        it is 0 then connections may remain in the pool indefinitely.
        """
        return self._impl.max_lifetime_session

    @property
    def max_sessions_per_shard(self) -> int:
        """
        The maximum number of connections that may be associated with a
        particular shard.
        """
        return self._impl.max_sessions_per_shard

    @property
    def min(self) -> int:
        """
        The minimum number of connections the pool should contain.
        """
        return self._impl.min

    @property
    def ping_interval(self) -> int:
        """
        Length of time (in seconds) after which an unused connection in the
        pool will be a candidate for pinging when pool.acquire() is called. If
        the ping to the database indicates the connection is not alive a
        replacement connection will be returned by pool.acquire(). If
        ping_interval is a negative value the ping functionality will be
        disabled.
        """
        return self._impl.ping_interval

    @property
    def session_callback(self) -> Callable:
        """
        A callable that is invoked when a connection is returned from the pool
        for the first time, or when the connection tag differs from the one
        requested.
        """
        return self._impl.session_callback

    @property
    def soda_metadata_cache(self) -> bool:
        """
        Boolean indicating whether or not the SODA metadata cache should be
        enabled.
        """
        return self._impl.soda_metadata_cache

    @property
    def timeout(self) -> int:
        """
        Length of time (in seconds) that a connection may remain idle in the
        pool before it is terminated. If it is 0 then connections are never
        terminated.
        """
        return self._impl.timeout

    @property
    def wait_timeout(self) -> int:
        """
        Length of time (in milliseconds) that a caller should wait when
        acquiring a connection from the pool with getmode set to
        oracledb.POOL_GETMODE_TIMEDWAIT.
        """
        return self._impl.wait_timeout

    def copy(self) -> "PoolParams":
        """
        Creates a copy of the parameters and returns it.
        """
        params = PoolParams.__new__(PoolParams)
        params._impl = self._impl.copy()
        return params

    @utils.params_setter
    def set(self, *,
            min: int=None,
            max: int=None,
            increment: int=None,
            connectiontype: Type["oracledb.Connection"]=None,
            getmode: int=None,
            homogeneous: bool=None,
            timeout: int=None,
            wait_timeout: int=None,
            max_lifetime_session: int=None,
            session_callback: Callable=None,
            max_sessions_per_shard: int=None,
            soda_metadata_cache: bool=None,
            ping_interval: int=None,
            user: str=None,
            proxy_user: str=None,
            password: str=None,
            newpassword: str=None,
            wallet_password: str=None,
            access_token: Union[str, tuple, Callable]=None,
            host: str=None,
            port: int=None,
            protocol: str=None,
            https_proxy: str=None,
            https_proxy_port: int=None,
            service_name: str=None,
            sid: str=None,
            server_type: str=None,
            cclass: str=None,
            purity: int=None,
            expire_time: int=None,
            retry_count: int=None,
            retry_delay: int=None,
            tcp_connect_timeout: float=None,
            ssl_server_dn_match: bool=None,
            ssl_server_cert_dn: str=None,
            wallet_location: str=None,
            events: bool=None,
            externalauth: bool=None,
            mode: int=None,
            disable_oob: bool=None,
            stmtcachesize: int=None,
            edition: str=None,
            tag: str=None,
            matchanytag: bool=None,
            config_dir: str=None,
            appcontext: list=None,
            shardingkey: list=None,
            supershardingkey: list=None,
            debug_jdwp: str=None,
            handle: int=None,
            threaded: bool=None,
            encoding: str=None,
            nencoding: str=None,
            waitTimeout: int=None,
            maxLifetimeSession: int=None,
            maxSessionsPerShard: int=None,
            sessionCallback: Callable=None
           ):
        """
        All parameters are optional. A brief description of each parameter
        follows:

        - min: the minimum number of connections the pool should contain

        - max: the maximum number of connections the pool should contain

        - increment: the number of connections that should be added to the pool
          whenever a new connection needs to be created

        - connectiontype: the class of the connection that should be returned
          during calls to pool.acquire(). It must be Connection or a subclass
          of Connection

        - getmode: how pool.acquire() will behave. One of the constants
          oracledb.POOL_GETMODE_WAIT, oracledb.POOL_GETMODE_NOWAIT,
          oracledb.POOL_GETMODE_FORCEGET, or oracledb.POOL_GETMODE_TIMEDWAIT

        - homogeneous: a boolean indicating whether the connections are
          homogeneous (same user) or heterogeneous (multiple users)

        - timeout: length of time (in seconds) that a connection may remain
          idle in the pool before it is terminated. If it is 0 then connections
          are never terminated

        - wait_timeout: length of time (in milliseconds) that a caller should
          wait when acquiring a connection from the pool with getmode set to
          oracledb.POOL_GETMODE_TIMEDWAIT

        - max_lifetime_session: length of time (in seconds) that connections
          can remain in the pool. If it is 0 then connections may remain in the
          pool indefinitely

        - session_callback: a callable that is invoked when a connection is
          returned from the pool for the first time, or when the connection tag
          differs from the one requested

        - max_sessions_per_shard: the maximum number of connections that may be
          associated with a particular shard

        - soda_metadata_cache: boolean indicating whether or not the SODA
          metadata cache should be enabled

        - ping_interval: length of time (in seconds) after which an unused
          connection in the pool will be a candidate for pinging when
          pool.acquire() is called. If the ping to the database indicates the
          connection is not alive a replacement connection will be returned by
          pool.acquire(). If ping_interval is a negative value the ping
          functionality will be disabled

        - user: the name of the user to connect to

        - proxy_user: the name of the proxy user to connect to. If this value
          is not specified, it will be parsed out of user if user is in the
          form "user[proxy_user]"

        - password: the password for the user

        - newpassword: the new password for the user. The new password will
          take effect immediately upon a successful connection to the database

        - wallet_password: the password to use to decrypt the wallet, if it is
          encrypted. This value is only used in thin mode

        - access_token: expected to be a string or a 2-tuple or a callable. If
          it is a string, it specifies an Azure AD OAuth2 token used for Open
          Authorization (OAuth 2.0) token based authentication. If it is a
          2-tuple, it specifies the token and private key strings used for
          Oracle Cloud Infrastructure (OCI) Identity and Access Management
          (IAM) token based authentication. If it is a callable, it returns
          either a string or a 2-tuple used for OAuth 2.0 or OCI IAM token
          based authentication and is useful when the pool needs to expand and
          create new connections but the current authentication token has
          expired

        - host: the name or IP address of the machine hosting the database or
          the database listener

        - port: the port number on which the database listener is listening

        - protocol: one of the strings "tcp" or "tcps" indicating whether to
          use unencrypted network traffic or encrypted network traffic (TLS)

        - https_proxy: the name or IP address of a proxy host to use for
          tunneling secure connections

        - https_proxy_port: the port on which to communicate with the proxy
          host

        - service_name: the service name of the database

        - sid: the system identifier (SID) of the database. Note using a
          service_name instead is recommended

        - server_type: the type of server connection that should be
          established. If specified, it should be one of "dedicated", "shared"
          or "pooled"

        - cclass: connection class to use for Database Resident Connection
          Pooling (DRCP)

        - purity: purity to use for Database Resident Connection Pooling (DRCP)

        - expire_time: an integer indicating the number of minutes between the
          sending of keepalive probes. If this parameter is set to a value
          greater than zero it enables keepalive

        - retry_count: the number of times that a connection attempt should be
          retried before the attempt is terminated

        - retry_delay: the number of seconds to wait before making a new
          connection attempt

        - tcp_connect_timeout: a float indicating the maximum number of seconds
          to wait for establishing a connection to the database host

        - ssl_server_dn_match: boolean indicating whether the server
          certificate distinguished name (DN) should be matched in addition to
          the regular certificate verification that is performed. Note that if
          the ssl_server_cert_dn parameter is not privided, host name matching
          is performed instead

        - ssl_server_cert_dn: the distinguished name (DN) which should be
          matched with the server. This value is ignored if the
          ssl_server_dn_match parameter is not set to the value True. If
          specified this value is used for any verfication. Otherwise the
          hostname will be used.

        - wallet_location: the directory where the wallet can be found. In thin
          mode this must be the directory containing the PEM-encoded wallet
          file ewallet.pem. In thick mode this must be the directory containing
          the file cwallet.sso

        - events: boolean specifying whether events mode should be enabled.
          This value is only used in thick mode and is needed for continuous
          query notification and high availability event notifications

        - externalauth: a boolean indicating whether to use external
          authentication

        - mode: authorization mode to use. For example
          oracledb.AUTH_MODE_SYSDBA

        - disable_oob: boolean indicating whether out-of-band breaks should be
          disabled. This value is only used in thin mode. It has no effect on
          Windows which does not support this functionality

        - stmtcachesize: identifies the initial size of the statement cache

        - edition: edition to use for the connection. This parameter cannot be
          used simultaneously with the cclass parameter

        - tag: identifies the type of connection that should be returned from a
          pool. This value is only used in thick mode

        - matchanytag: boolean specifying whether any tag can be used when
          acquiring a connection from the pool. This value is only used in
          thick mode.

        - config_dir: directory in which the optional tnsnames.ora
          configuration file is located. This value is only used in thin mode.
          For thick mode use the config_dir parameter of init_oracle_client()

        - appcontext: application context used by the connection. It should be
          a list of 3-tuples (namespace, name, value) and each entry in the
          tuple should be a string. This value is only used in thick mode

        - shardingkey: a list of strings, numbers, bytes or dates that identify
          the database shard to connect to. This value is only used in thick
          mode

        - supershardingkey: a list of strings, numbers, bytes or dates that
          identify the database shard to connect to. This value is only used in
          thick mode

        - debug_jdwp: a string with the format "host=<host>;port=<port>" that
          specifies the host and port of the PL/SQL debugger. This value is
          only used in thin mode. For thick mode set the ORA_DEBUG_JDWP
          environment variable

        - handle: an integer representing a pointer to a valid service context
          handle. This value is only used in thick mode. It should be used with
          extreme caution
        """
        pass



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
2700 - Module for testing AQ
"""

import datetime
import decimal
import threading
import unittest

import oracledb
import test_env

@unittest.skipIf(test_env.get_is_thin(), "thin mode doesn't support AQ yet")
class TestCase(test_env.BaseTestCase):
    book_type_name = "UDT_BOOK"
    book_queue_name = "TEST_BOOK_QUEUE"
    book_data = [
        ("Wings of Fire", "A.P.J. Abdul Kalam", decimal.Decimal("15.75")),
        ("The Story of My Life", "Hellen Keller", decimal.Decimal("10.50")),
        ("The Chronicles of Narnia", "C.S. Lewis", decimal.Decimal("25.25"))
    ]
    json_queue_name = "TEST_JSON_QUEUE"
    json_data = [
        [
            2.75,
            True,
            'Ocean Beach',
            b'Some bytes',
            {'keyA': 1.0, 'KeyB': 'Melbourne'},
            datetime.datetime(2022, 8, 1, 0, 0)
        ],
        [
            True,
            False,
            'String',
            b'Some Bytes',
            {},
            {"name": None},
            {"name": "John"},
            {"age": 30},
            {"Permanent": True},
            {
                "employee": {
                    "name":"John",
                    "age": 30,
                    "city": "Delhi",
                    "Parmanent": True
                }
            },
            {
                "employees": ["John", "Matthew", "James"]
            },
            {
                "employees": [
                    {
                        "employee1": {"name": "John", "city": "Delhi"}
                    },
                    {
                        "employee2": {"name": "Matthew", "city": "Mumbai"}
                    },
                    {
                        "employee3": {"name": "James", "city": "Bangalore"}
                    }
                ]
            }
        ],
        [
            datetime.datetime.today(),
            datetime.datetime(2004, 2, 1, 3, 4, 5),
            datetime.datetime(2020, 12, 2, 13, 29, 14),
            datetime.timedelta(8.5),
            datetime.datetime(2002, 12, 13, 9, 36, 0),
            oracledb.Timestamp(2002, 12, 13, 9, 36, 0),
            datetime.datetime(2002, 12, 13)
        ],
        dict(name="John", age=30, city="New York"),
        [
            0,
            1,
            25.25,
            6088343244,
            -9999999999999999999,
            decimal.Decimal("0.25"),
            decimal.Decimal("10.25"),
            decimal.Decimal("319438950232418390.273596")
        ]
    ]

    def __deq_in_thread(self, results):
        with test_env.get_connection() as connection:
            books_type = connection.gettype(self.book_type_name)
            queue = connection.queue(self.book_queue_name, books_type)
            queue.deqoptions.wait = 10
            props = queue.deqone()
            if props is not None:
                book = props.payload
                results.append((book.TITLE, book.AUTHORS, book.PRICE))
            connection.commit()

    def __verify_attr(self, obj, attrName, value):
        setattr(obj, attrName, value)
        self.assertEqual(getattr(obj, attrName), value)

    def test_2700_deq_empty(self):
        "2700 - test dequeuing an empty queue"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        self.assertTrue(props is None)

    def test_2701_deq_enq(self):
        "2701 - test enqueuing and dequeuing multiple messages"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        props = self.connection.msgproperties()
        for title, authors, price in self.book_data:
            props.payload = book = queue.payload_type.newobject()
            book.TITLE = title
            book.AUTHORS = authors
            book.PRICE = price
            queue.enqone(props)
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        results = []
        while True:
            props = queue.deqone()
            if props is None:
                break
            book = props.payload
            row = (book.TITLE, book.AUTHORS, book.PRICE)
            results.append(row)
        self.connection.commit()
        self.assertEqual(results, self.book_data)

    def test_2702_deq_mode_remove_no_data(self):
        "2702 - test dequeuing with DEQ_REMOVE_NODATA option"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        book = queue.payload_type.newobject()
        title, authors, price = self.book_data[1]
        book.TITLE = title
        book.AUTHORS = authors
        book.PRICE = price
        props = self.connection.msgproperties(payload=book)
        queue.enqone(props)
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.mode = oracledb.DEQ_REMOVE_NODATA
        props = queue.deqone()
        self.assertTrue(props is not None)
        self.assertTrue(props.payload.TITLE is None)

    def test_2703_deq_options(self):
        "2703 - test getting/setting dequeue options attributes"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        options = queue.deqoptions
        self.__verify_attr(options, "condition", "TEST_CONDITION")
        self.__verify_attr(options, "consumername", "TEST_CONSUMERNAME")
        self.__verify_attr(options, "correlation", "TEST_CORRELATION")
        self.__verify_attr(options, "mode", oracledb.DEQ_LOCKED)
        self.__verify_attr(options, "navigation",
                           oracledb.DEQ_NEXT_TRANSACTION)
        self.__verify_attr(options, "transformation", "TEST_TRANSFORMATION")
        self.__verify_attr(options, "visibility", oracledb.ENQ_IMMEDIATE)
        self.__verify_attr(options, "wait", 1287)
        self.__verify_attr(options, "msgid", b'mID')

    def test_2704_deq_with_wait(self):
        "2704 - test waiting for dequeue"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        results = []
        thread = threading.Thread(target=self.__deq_in_thread, args=(results,))
        thread.start()
        book = queue.payload_type.newobject()
        title, authors, price = self.book_data[0]
        book.TITLE = title
        book.AUTHORS = authors
        book.PRICE = price
        props = self.connection.msgproperties(payload=book)
        queue.enqone(props)
        self.connection.commit()
        thread.join()
        self.assertEqual(results, [(title, authors, price)])

    def test_2705_enq_options(self):
        "2705 - test getting/setting enqueue options attributes"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        options = queue.enqoptions
        self.__verify_attr(options, "visibility", oracledb.ENQ_IMMEDIATE)

    def test_2706_errors_for_invalid_values(self):
        "2706 - test errors for invalid values for enqueue"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        book = queue.payload_type.newobject()
        self.assertRaises(TypeError, queue.enqone, book)

    def test_2707_msg_props(self):
        "2707 - test getting/setting message properties attributes"
        props = self.connection.msgproperties()
        self.__verify_attr(props, "correlation", "TEST_CORRELATION")
        self.__verify_attr(props, "delay", 60)
        self.__verify_attr(props, "exceptionq", "TEST_EXCEPTIONQ")
        self.__verify_attr(props, "expiration", 30)
        self.assertEqual(props.attempts, 0)
        self.__verify_attr(props, "priority", 1)
        self.assertEqual(props.state, oracledb.MSG_READY)
        self.assertEqual(props.deliverymode, 0)

    def test_2708_visibility_mode_commit(self):
        "2708 - test enqueue visibility option - ENQ_ON_COMMIT"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.visibility = oracledb.ENQ_ON_COMMIT
        props = self.connection.msgproperties(payload=book)
        queue.enqone(props)

        other_connection = test_env.get_connection()
        books_type = other_connection.gettype(self.book_type_name)
        queue = other_connection.queue(self.book_queue_name, books_type)
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        self.assertTrue(props is None)
        self.connection.commit()
        props = queue.deqone()
        self.assertTrue(props is not None)

    def test_2709_visibility_mode_immediate(self):
        "2709 - test enqueue visibility option - ENQ_IMMEDIATE"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.connection.msgproperties(payload=book)
        queue.enqone(props)

        other_connection = test_env.get_connection()
        books_type = other_connection.gettype(self.book_type_name)
        queue = other_connection.queue(self.book_queue_name, books_type)
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_ON_COMMIT
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        book = props.payload
        results = (book.TITLE, book.AUTHORS, book.PRICE)
        other_connection.commit()
        self.assertEqual(results, self.book_data[0])

    def test_2710_delivery_mode_same_buffered(self):
        "2710 - test enqueue/dequeue delivery modes identical - buffered"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_BUFFERED
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.connection.msgproperties(payload=book)
        queue.enqone(props)

        other_connection = test_env.get_connection()
        books_type = other_connection.gettype(self.book_type_name)
        queue = other_connection.queue(self.book_queue_name, books_type)
        queue.deqoptions.deliverymode = oracledb.MSG_BUFFERED
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        book = props.payload
        results = (book.TITLE, book.AUTHORS, book.PRICE)
        other_connection.commit()
        self.assertEqual(results, self.book_data[0])

    def test_2711_delivery_mode_same_persistent(self):
        "2711 - test enqueue/dequeue delivery modes identical - persistent"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_PERSISTENT
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.connection.msgproperties(payload=book)
        queue.enqone(props)

        other_connection = test_env.get_connection()
        books_type = other_connection.gettype(self.book_type_name)
        queue = other_connection.queue(self.book_queue_name, books_type)
        queue.deqoptions.deliverymode = oracledb.MSG_PERSISTENT
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        book = props.payload
        results = (book.TITLE, book.AUTHORS, book.PRICE)
        other_connection.commit()
        self.assertEqual(results, self.book_data[0])

    def test_2712_delivery_mode_same_persistent_buffered(self):
        "2712 - test enqueue/dequeue delivery modes the same"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_PERSISTENT_OR_BUFFERED
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.connection.msgproperties(payload=book)
        queue.enqone(props)

        other_connection = test_env.get_connection()
        books_type = other_connection.gettype(self.book_type_name)
        queue = other_connection.queue(self.book_queue_name, books_type)
        queue.deqoptions.deliverymode = oracledb.MSG_PERSISTENT_OR_BUFFERED
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        book = props.payload
        results = (book.TITLE, book.AUTHORS, book.PRICE)
        other_connection.commit()
        self.assertEqual(results, self.book_data[0])

    def test_2713_delivery_mode_different(self):
        "2713 - test enqueue/dequeue delivery modes different"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue.enqoptions.deliverymode = oracledb.MSG_BUFFERED
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        props = self.connection.msgproperties(payload=book)
        queue.enqone(props)

        other_connection = test_env.get_connection()
        books_type = other_connection.gettype(self.book_type_name)
        queue = other_connection.queue(self.book_queue_name, books_type)
        queue.deqoptions.deliverymode = oracledb.MSG_PERSISTENT
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        self.assertTrue(props is None)

    def test_2714_dequeue_transformation(self):
        "2714 - test dequeue transformation"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        expected_price = book.PRICE + 10
        props = self.connection.msgproperties(payload=book)
        queue.enqone(props)
        self.connection.commit()

        other_connection = test_env.get_connection()
        books_type = other_connection.gettype(self.book_type_name)
        queue = other_connection.queue(self.book_queue_name, books_type)
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        transformation_str = "%s.transform2" % self.connection.username
        queue.deqoptions.transformation = transformation_str
        self.assertEqual(queue.deqOptions.transformation, transformation_str)
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        self.assertEqual(props.payload.PRICE, expected_price)

    def test_2715_enqueue_transformation(self):
        "2715 - test enqueue transformation"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        expected_price = book.PRICE + 5
        queue.enqoptions.transformation = \
        transformation_str = "%s.transform1" % self.connection.username
        queue.enqoptions.transformation = transformation_str
        props = self.connection.msgproperties(payload=book)
        queue.enqone(props)
        self.connection.commit()

        other_connection = test_env.get_connection()
        books_type = other_connection.gettype(self.book_type_name)
        queue = other_connection.queue(self.book_queue_name, books_type)
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        props = queue.deqone()
        self.assertEqual(props.payload.PRICE, expected_price)

    def test_2716_payloadType_deprecation(self):
        "2716 - test to verify payloadType is deprecated"
        books_type = self.connection.gettype(self.book_type_name)
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2014:",
                               self.connection.queue, self.book_queue_name,
                               books_type, payloadType=books_type)

    def test_2717_message_with_no_payload(self):
        "2717 - test error for message with no payload"
        books_type = self.connection.gettype(self.book_type_name)
        queue = self.connection.queue(self.book_queue_name, books_type)
        props = self.connection.msgproperties()
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2000:",
                               queue.enqone, props)

    def test_2718_verify_msgid(self):
        "2718 - verify that the msgid property is returned correctly"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        props = self.connection.msgproperties(payload=book)
        self.assertEqual(props.msgid, None)
        queue.enqone(props)
        self.cursor.execute("select msgid from book_queue_tab")
        actual_msgid, = self.cursor.fetchone()
        self.assertEqual(props.msgid, actual_msgid)
        props = queue.deqone()
        self.assertEqual(props.msgid, actual_msgid)

    def test_2719_recipients_list(self):
        "2719 - verify use of recipients property"
        books_type = self.connection.gettype(self.book_type_name)
        book = books_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        queue = self.connection.queue("BOOK_QUEUE_MULTI", books_type)
        props = self.connection.msgproperties(payload=book,
                                              recipients=["sub2", "sub3"])
        self.assertEqual(props.recipients, ["sub2", "sub3"])
        queue.enqone(props)
        self.connection.commit()
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
        queue.deqoptions.consumername = "sub3"
        props1 = queue.deqone()
        book = props1.payload
        results = (book.TITLE, book.AUTHORS, book.PRICE)
        self.assertEqual(results, self.book_data[0])
        queue.deqoptions.consumername = "sub1"
        props1 = queue.deqone()
        self.assertTrue(props1 is None)

    def test_2720_aq_notification(self):
        "2720 - verify msgid of aq message which spawned notification "
        if self.is_on_oracle_cloud(self.connection):
            self.skipTest("AQ notification not supported on the cloud")
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        condition = threading.Condition()
        connection = test_env.get_connection(events=True)
        def notification_callback(message):
            self.cursor.execute("select msgid from book_queue_tab")
            actual_msgid, = self.cursor.fetchone()
            self.assertEqual(actual_msgid, message.msgid)
            with condition:
                condition.notify()
        sub = connection.subscribe(namespace=oracledb.SUBSCR_NAMESPACE_AQ,
                                   name=self.book_queue_name,
                                   callback=notification_callback, timeout=300)
        book = queue.payload_type.newobject()
        book.TITLE, book.AUTHORS, book.PRICE = self.book_data[0]
        props = self.connection.msgproperties(payload=book)
        queue.enqone(props)
        self.connection.commit()
        with condition:
            self.assertTrue(condition.wait(5))

    def test_2721_json_enq_deq(self):
        "2721 - test enqueuing and dequeuing JSON payloads"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        self.assertEqual(queue.payload_type, "JSON")
        for data in self.json_data:
            props = self.connection.msgproperties(payload=data)
            queue.enqone(props)
        self.connection.commit()
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        results = []
        while True:
            props = queue.deqone()
            if props is None:
                break
            results.append(props.payload)
        self.connection.commit()
        self.assertEqual(results, self.json_data)

    def test_2722_no_json_payload(self):
        "2722 - test enqueuing to a JSON queue without a JSON payload"
        queue = self.get_and_clear_queue(self.json_queue_name, "JSON")
        random_string = "This is a string message"
        props = self.connection.msgproperties(payload=random_string)
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPI-1071:",
                               queue.enqone, props)

    def test_2723_enqtime(self):
        "2723 - test message props enqtime"
        queue = self.get_and_clear_queue(self.book_queue_name,
                                         self.book_type_name)
        book = queue.payload_type.newobject()
        self.cursor.execute("select sysdate from dual")
        start_date, = self.cursor.fetchone()
        props = self.connection.msgproperties(payload=book)
        queue.enqone(props)
        props = queue.deqone()
        self.cursor.execute("select sysdate from dual")
        end_date, = self.cursor.fetchone()
        self.assertTrue(start_date <= props.enqtime <= end_date)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# queue.pyx
#
# Cython file defining the base Queue implementation class (embedded in
# base_impl.pyx).
#------------------------------------------------------------------------------

cdef class BaseQueueImpl:

    @utils.CheckImpls("dequeuing multiple messages")
    def deq_many(self, uint32_t max_num_messages):
        pass

    @utils.CheckImpls("dequeuing a single message")
    def deq_one(self):
        pass

    @utils.CheckImpls("enqueueing multiple messages")
    def enq_many(self, list props_impls):
        pass

    @utils.CheckImpls("enqueuing a single message")
    def enq_one(self, BaseMsgPropsImpl props_impl):
        pass

    @utils.CheckImpls("initializing a queue")
    def initialize(self, BaseConnImpl conn_impl, str name,
                   BaseDbObjectImpl payload_type, bint is_json):
        pass


cdef class BaseDeqOptionsImpl:

    @utils.CheckImpls("getting the condition")
    def get_condition(self):
        pass

    @utils.CheckImpls("getting the consumer name")
    def get_consumer_name(self):
        pass

    @utils.CheckImpls("getting the correlation")
    def get_correlation(self):
        pass

    @utils.CheckImpls("getting the message id")
    def get_message_id(self):
        pass

    @utils.CheckImpls("getting the mode")
    def get_mode(self):
        pass

    @utils.CheckImpls("getting the navigation")
    def get_navigation(self):
        pass

    @utils.CheckImpls("getting the transformation")
    def get_transformation(self):
        pass

    @utils.CheckImpls("getting the visibility")
    def get_visibility(self):
        pass

    @utils.CheckImpls("getting the wait time")
    def get_wait(self):
        pass

    @utils.CheckImpls("setting the condition")
    def set_condition(self, str value):
        pass

    @utils.CheckImpls("setting the consumer name")
    def set_consumer_name(self, str value):
        pass

    @utils.CheckImpls("setting the correlation")
    def set_correlation(self, str value):
        pass

    @utils.CheckImpls("setting the delivery mode")
    def set_delivery_mode(self, uint16_t value):
        pass

    @utils.CheckImpls("setting the mode")
    def set_mode(self, uint32_t value):
        pass

    @utils.CheckImpls("setting the message id")
    def set_message_id(self, bytes value):
        pass

    @utils.CheckImpls("setting the navigation")
    def set_navigation(self, uint32_t value):
        pass

    @utils.CheckImpls("setting the transformation")
    def set_transformation(self, str value):
        pass

    @utils.CheckImpls("setting the visibility")
    def set_visibility(self, uint32_t value):
        pass

    @utils.CheckImpls("setting the wait time")
    def set_wait(self, uint32_t value):
        pass


cdef class BaseEnqOptionsImpl:

    @utils.CheckImpls("getting the transformation")
    def get_transformation(self):
        pass

    @utils.CheckImpls("getting the visibility")
    def get_visibility(self):
        pass

    @utils.CheckImpls("setting the delivery mode")
    def set_delivery_mode(self, uint16_t value):
        pass

    @utils.CheckImpls("setting the transformation")
    def set_transformation(self, str value):
        pass

    @utils.CheckImpls("setting the visibility")
    def set_visibility(self, uint32_t value):
        pass


cdef class BaseMsgPropsImpl:

    @utils.CheckImpls("getting the number of attempts")
    def get_num_attempts(self):
        pass

    @utils.CheckImpls("getting the correlation")
    def get_correlation(self):
        pass

    @utils.CheckImpls("getting the delay")
    def get_delay(self):
        pass

    @utils.CheckImpls("getting the delivery mode")
    def get_delivery_mode(self):
        pass

    @utils.CheckImpls("getting the enqueue time")
    def get_enq_time(self):
        pass

    @utils.CheckImpls("getting the name of the exception queue")
    def get_exception_queue(self):
        pass

    @utils.CheckImpls("getting the expiration")
    def get_expiration(self):
        pass

    @utils.CheckImpls("getting the message id")
    def get_message_id(self):
        pass

    @utils.CheckImpls("getting the priority")
    def get_priority(self):
        pass

    @utils.CheckImpls("getting the message state")
    def get_state(self):
        pass

    @utils.CheckImpls("setting the correlation")
    def set_correlation(self, str value):
        pass

    @utils.CheckImpls("setting the delay")
    def set_delay(self, int32_t value):
        pass

    @utils.CheckImpls("setting the name of the exception queue")
    def set_exception_queue(self, str value):
        pass

    @utils.CheckImpls("setting the expiration")
    def set_expiration(self, int32_t value):
        pass

    @utils.CheckImpls("setting the payload from bytes")
    def set_payload_bytes(self, bytes value):
        pass

    @utils.CheckImpls("setting the payload from a database object")
    def set_payload_object(self, BaseDbObjectImpl value):
        pass

    @utils.CheckImpls("setting the priority")
    def set_priority(self, int32_t value):
        pass

    @utils.CheckImpls("setting recipients list")
    def set_recipients(self, list value):
        pass



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

import os
import platform
import sys
import sysconfig
from setuptools import setup, Extension

# base source directory
source_dir = os.path.join("src", "oracledb")

# determine the base implementation dependent source files (included)
impl_dir = os.path.join(source_dir, "impl", "base")
base_depends = [os.path.join(impl_dir, n) \
                for n in sorted(os.listdir(impl_dir)) if n.endswith(".pyx")]
base_pxd = os.path.join(source_dir, "base_impl.pxd")
base_depends.append(base_pxd)

# determine the thick mode dependent source files (included)
impl_dir = os.path.join(source_dir, "impl", "thick")
source_dirs = [
        os.path.join(impl_dir, "odpi", "src"),
        os.path.join(impl_dir, "odpi", "include")
]
thick_depends = [os.path.join(d, n) \
                 for d in source_dirs
                 for n in sorted(os.listdir(d)) \
                 if n.endswith(".c") or n.endswith(".h")]
thick_depends.extend(os.path.join(impl_dir, n) \
                     for n in sorted(os.listdir(impl_dir))
                     if n.endswith(".pyx") or n.endswith(".pxd"))
thick_depends.append(base_pxd)

# determine the thin mode dependent source files (included)
impl_dir = os.path.join(source_dir, "impl", "thin")
thin_depends = [os.path.join(impl_dir, n) \
                for n in sorted(os.listdir(impl_dir))
                if n.endswith(".pyx") or n.endswith(".pxi")
                        or n.endswith(".pxd")]
thin_depends.append(base_pxd)

# if the platform is macOS:
#  - target the minimim OS version that current Python packages work with.
#    (Use 'otool -l /path/to/python' and look for 'version' in the
#    LC_VERSION_MIN_MACOSX section)
#  - add argument required for cross-compilation for both x86_64 and arm64
#    architectures if the python interpreter is a universal2 version.

extra_compile_args = []
if sys.platform == "darwin":
    extra_compile_args.extend(["-mmacosx-version-min=10.9"])
    if "universal2" in sysconfig.get_platform():
        if platform.machine() == "x86_64":
            target = "arm64-apple-macos"
        else:
            target = "x86_64-apple-macos"
        extra_compile_args.extend(["-target", target])

setup(
    ext_modules=[
        Extension("oracledb.base_impl",
                  sources=["src/oracledb/base_impl.pyx"],
                  depends=base_depends,
                  extra_compile_args=extra_compile_args),
        Extension("oracledb.thin_impl",
                  sources=["src/oracledb/thin_impl.pyx"],
                  depends=thin_depends,
                  extra_compile_args=extra_compile_args),
        Extension("oracledb.thick_impl",
                  sources=["src/oracledb/thick_impl.pyx"],
                  include_dirs=["src/oracledb/impl/thick/odpi/include"],
                  depends=thick_depends,
                  extra_compile_args=extra_compile_args)
    ]
)
